/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "abiword2usfm.h"
#include "utilities.h"
#include "gwrappers.h"
#include "abiwordutils.h"
#include "usfmtools.h"


Abiword2Usfm::Abiword2Usfm (const vector<ustring>& lines, vector<Style> * styles)
{
  // Initialize variables
  mystyles = styles;
  in_footnote_element = false;
  in_non_relevant_element = false;
  in_paragraph_or_character_element = 0;
  
  // Get contents.
  ustring line;
  for (unsigned int i = 0; i < lines.size(); i++) {
    line.append (lines[i]);
    line.append ("\n");
  }
  char *contents = strdup (line.c_str());
  
  // Create input buffer.  
  xmlParserInputBufferPtr inputbuffer;
  inputbuffer = xmlParserInputBufferCreateMem(contents, strlen (contents), XML_CHAR_ENCODING_NONE);
  
  // Iterate through the nodes in streaming mode.
  success = true;
  xmlTextReaderPtr reader;
  reader = xmlNewTextReader (inputbuffer, NULL);
  if (reader) {
    int read;
    while ((read = xmlTextReaderRead(reader) == 1)) {
      switch (xmlTextReaderNodeType (reader)) {
        case XML_READER_TYPE_ELEMENT:
        {
          start_element_handler (reader);
          // If the node is empty, e.g. "<a/>", call the endhandler too.
          if (xmlTextReaderIsEmptyElement (reader)) {
            end_element_handler (reader);
          }
          break;
        }
        case XML_READER_TYPE_TEXT:
        {
          text_handler (reader);
          break;
        }
        case XML_READER_TYPE_END_ELEMENT:
        {
          end_element_handler (reader);
          break;
        }
      }
    }
    if (read != 0) {
      success = false;
    }
  } else {
    success = false;
  }
  
  // Free some memory.
  if (reader) xmlFreeTextReader(reader);
  if (inputbuffer) xmlFreeParserInputBuffer (inputbuffer);
  if (contents) free (contents);  
}


Abiword2Usfm::~Abiword2Usfm ()
{
}


void Abiword2Usfm::start_element_handler (xmlTextReaderPtr reader)
// When we encounter a new element that starts data, this handler deals with that.
{
  // Categorize the element that we enter.
  categorize_element (reader, true);

  // Bail out if in a non-relevant element.
  if (in_non_relevant_element) return;
  if (!in_paragraph_or_character_element) return;
  
  // If we're in a footnote, bail out.
  if (in_footnote_element) 
    return;

  // Get the style attribute of this node, and store it.
  // The style is the attribute value of attribute name "style".
  ustring current_style;
  xmlChar * attribute;
  attribute = xmlTextReaderGetAttribute (reader, (const xmlChar *) "style");
  if (attribute) 
    current_style = (const char*) attribute;
  xmlFree(attribute);
  if (current_style == abiword_unknown_style ())
    current_style.clear();
  current_styles.push_back (current_style);

  // Get some relevant properties of this style.
  bool start_newline, showmarker, plaintext, b;
  style_get_properties (current_style, start_newline, b, showmarker, plaintext, b);
  // If this marker starts a new line in the usfm, treat it so, taking care not
  // to loose any data.
  if (start_newline) {
    outputline = trim (outputline);
    if (!outputline.empty()) output.push_back (outputline);
    outputline.clear();
    // If abiword hid this marker, add it again.
    if (!showmarker) {
      if (!current_style.empty()) {
        outputline.append (usfm_get_full_opening_marker (current_style));
      }
    }
  } else {
    // Treatment of inline markers.
    if (!current_style.empty()) {
      outputline.append (usfm_get_full_opening_marker (current_style));
    }
  }
}


void Abiword2Usfm::end_element_handler (xmlTextReaderPtr reader)
// When we encounter an element that ends data, this handler deals with that.
{
  // Only proceed if we are in a relevant element.
  if (!in_non_relevant_element && (in_paragraph_or_character_element)) {

    // Handle footnote stuff.
    if (in_footnote_element) {
      outputline.append (trim (footnoteline));
      footnoteline.clear();
    }
    
    // Handle normal text.
    else {

      // Get the marker of the style that is now closing.
      ustring marker = current_style ();  

      // Get properties of this style.
      bool newline_on_end, has_endmarker, b;
      style_get_properties (marker, b, newline_on_end, b, b, has_endmarker);

      // Add endmarker if needed.
      if (has_endmarker) {
        if (!marker.empty()) {
          outputline.append (usfm_get_full_closing_marker (marker));
        }
      }
  
      // If a new line should be started when this marker finishes, handle that.
      if (newline_on_end) {
        outputline = trim (outputline);
        if (!outputline.empty()) output.push_back (outputline);
        outputline.clear();
      }
  
      // Remove the last style from the stack.
      if (!current_styles.empty())
        current_styles.pop_back ();
      
    }
    
  }
  
  // Categorize the element that we enter.
  categorize_element (reader, false);
}


void Abiword2Usfm::text_handler (xmlTextReaderPtr reader)
// When we encounter the text inside an element, this handler deals with that.
{
  // Bail out if we are not in a relevant element.
  if (in_non_relevant_element) return;
  if (!in_paragraph_or_character_element) return;

  // Get text.
  ustring text;
  
  xmlChar * value;
  value = xmlTextReaderValue(reader);
  if (value) {
    text = (const char *) value;
    xmlFree(value);
  }
  
  // Handle footnote case.
  if (in_footnote_element) {
    footnoteline.append (text);
  }
  // Handle normal text case.
  else {
    // Add the text to the outputline we're building.
    outputline.append (text);      
  }
}


ustring Abiword2Usfm::current_style ()
// Retrieves the current style. This is the style on top of the stack.
{
  ustring style;
  if (!current_styles.empty()) {
    style = current_styles[current_styles.size() - 1];
  }
  return style;
}


void Abiword2Usfm::style_get_properties (const ustring& marker, bool& start_newline, bool& newline_on_end, bool& showmarker, bool& plaintext, bool& has_endmarker)
/*
Gets relevant properties of a style.
marker: the style to consider.
start_newline: returns whether this usfm marker should start a new line.
newline_on_end: whether a new line is started when this marker ends.
hidemarker: returns whether this usfm marker is hidden in the text.
plaintext: returns whether this style was put in plain text.
has_endmarkere: returns whether an endmarker should be written when this style ends.
*/
{
  // Find a Style for that marker, or NULL if it isn't there.
  Style * style = NULL;
  for (unsigned int i = 0; i < mystyles->size(); i++) {
    if (mystyles->at(i).marker == marker) {
      style = &mystyles->at(i);
      break;
    }
  }
  
  // Obtain the type and the subtype.
  StyleType style_type = (StyleType) G_MAXINT;
  int style_subtype = -1;
  if (style) {
    style_type = style->type;
    style_subtype = style->subtype;
  }
  
  // Produce flags whether it is a paragraph, and so forth.
  AbiwordStyleType abiwordstyletype;
  abiword_style_get_properties (style_type, style_subtype, abiwordstyletype, plaintext, showmarker);
  
  // Get our relevant properties.
  switch (abiwordstyletype) {
    case abwstParagraph:
    {
      start_newline = true;
      newline_on_end = true;
      has_endmarker = false;
      break;
    }
    case abwstCharacter:
    {
      start_newline = false;
      newline_on_end = false;
      has_endmarker = true;
      break;
    }
    case abwstVerse:
    {
      start_newline = true;
      newline_on_end = false;
      has_endmarker = false;
      break;
    }
    case abwstFootnote:
    {
      start_newline = false;
      newline_on_end = false;
      has_endmarker = true;
      break;
    }
    case abwstEndnote:
    {
      start_newline = false;
      newline_on_end = false;
      has_endmarker = true;
      break;
    }
  }
}


bool Abiword2Usfm::silently_skip_element (const xmlChar *element_name)
// Returns whether "element_name" does not influence our position tracking system.
{
  if (!xmlStrcmp(element_name, (const xmlChar *) "abiword")) return true;
  if (!xmlStrcmp(element_name, (const xmlChar *) "section")) return true;
  return false;
}


void Abiword2Usfm::categorize_element (xmlTextReaderPtr reader, bool start)
// This categorizes the element, that is, it sets or clears a number of variables.
// start: whether this is a start element.
{
  // Retrieve the name of the element.
  xmlChar * element_name;
  element_name = xmlTextReaderName(reader);

  // Note the area we enter.
  if (!xmlStrcmp(element_name, (const xmlChar *) "p")) {
    if (start) in_paragraph_or_character_element++;
    else in_paragraph_or_character_element--;
  }
  else if (!xmlStrcmp(element_name, (const xmlChar *) "c")) {
    if (start) in_paragraph_or_character_element++;
    else in_paragraph_or_character_element--;
  }
  else if (!xmlStrcmp(element_name, (const xmlChar *) "foot")) {
    if (start) in_footnote_element = true;
    else in_footnote_element = false;
  }
  else {
    if (!silently_skip_element (element_name)) {
      if (start) in_non_relevant_element = true;
      else in_non_relevant_element = false;
    }
  }

  // Free memory.
  if (element_name) xmlFree(element_name);
    
}
