/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "abiwordutils.h"
#include "utilities.h"
#include "constants.h"


void abiword_style_get_properties (StyleType type, int subtype, 
                                   AbiwordStyleType& abiwordstyletype, bool& plaintext, bool& showmarker)
/*
Get the properties of a certain style when displayed in Abiword.
abiwordstyletype: the type of this style, e.g. paragraph, foot/endnote, character style, etc.
plaintext: whether plain text is to be used.
showmarker: whether the marker should be displayed.
character: whether this marker marks a character style.
*/
{
  // Set default values, in case the value is not set below.
  abiwordstyletype = abwstParagraph;
  plaintext = true;
  showmarker = true;
  // Set values depending on the type of the marker, and the subtype.
  // Only set values if they differ from the defaults.
  switch (type) { // todo stTableElement not handled, fix it.
    case stIdentifier:
    case stNotUsedComment:
    case stNotUsedRunningHeader:
    {
      break;
    }
    case stStartsParagraph:
    {
      plaintext = false;
      showmarker = false;
      break;
    }
    case stInlineText:
    {
      abiwordstyletype = abwstCharacter;
      plaintext = false;
      showmarker = false;
      break;
    }
    case stChapterNumber:
    {
      plaintext = false;
      showmarker = false;
      break;
    }
    case stVerseNumber:
    {
      abiwordstyletype = abwstVerse;
      plaintext = false;
      showmarker = false;
      break;
    }
    case stFootEndNote:
    {
      abiwordstyletype = abwstFootnote;
      if ((subtype == fentContent) || (subtype == fentContentWithEndmarker)) {
        abiwordstyletype = abwstCharacter;
      }
      break;
    }
    case stCrossreference:
    {
      abiwordstyletype = abwstEndnote;
      if ((subtype == ctContent) || (subtype == ctContentWithEndmarker)) {
        abiwordstyletype = abwstCharacter;
      }
      break;
    }
    case stPeripheral:
    {
      break;
    }
    case stPicture:
    {
      break;
    }
    case stPageBreak:
    {
      break;
    }
  }
}


ustring abiword_style_produce_name (const ustring& marker)
// Produces the name of a style for abiword, e.g. "id".
{
  // It used to include round brackets for the marker, e.g. (id).
  // But for exporting to html, the brackets should be omitted.
  // Otherwise the styles do not show in the webbrows
  if (marker.empty()) return abiword_unknown_style ();
  return marker;
}


void abiword_style_add_property (bool& first_property_added, ustring& line, gchar * property, const ustring& value, gchar * suffix)
// When generating a style for abiword, this adds the right information for the 
// property to the line containing the style.
{
  if (first_property_added) {
    line.append ("; ");
  }
  first_property_added = true;
  line.append (property);
  line.append (":");
  line.append (value);
  line.append (suffix);
}


ustring abiword_unknown_style ()
// Gives the name of the style that is for markers that are not in the stylesheet.
{
  return "unknown";
}


/*

Abiword improvements needed.

If we were to create some styles for footnote, xref and endnote, would it
then be possible to change the style of a footnote? In the current abiword,
this does not work yet.

We might get good information if a signal is fired on xid change.
When creating a footnote, we could capture the xid number.
When the signal is fired and we are in this particular xid number, then
that would imply we are in that particular footnote.

There is a big disadvantage in abiword, when changing the style of a paragraph.
All characterstyles get lost when this is done. See how to fix that.

Each style has a "followedby" option that may come in useful later on to make
things quicker, e.g. a heading could be followed, on pressing return, by a paragraph.

How to handle a new paragraph in a footnote. Do it together with formatting the 
footnotes. Footnotes and endnotes and xrefs are now unformatted for reason that 
we have no possibility to change their styles.

*/
