/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "utilities.h"
#include <glib.h>
#include "editor.h"
#include "highlight.h"
#include "usfm.h"
#include "editornote.h"
#include <gdk/gdkkeysyms.h>
#include "projectutils.h"
#include "settings.h"
#include "screen.h"
#include "referenceutils.h"
#include "books.h"
#include "gwrappers.h"
#include "biblenotes.h"
#include "color.h"
#include "bible.h"


Editor::Editor (GtkWidget * notebook, const ustring& project_in):
current_reference (0, 1000, "")
{
  // Save variables.
  project = project_in;

  // Vertical box to put the Editor in.
  main_box = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (main_box);
  gtk_container_add (GTK_CONTAINER (notebook), main_box);

  // Label for this tab, and close button.
  GtkWidget *hbox9 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox9);
  gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook), -1), hbox9);
  
  GtkWidget * notebook_label = gtk_label_new (project_in.c_str());
  gtk_widget_show (notebook_label);
  gtk_box_pack_start (GTK_BOX (hbox9), notebook_label, FALSE, FALSE, 0);

  // Get the default height of the label.
  gint default_label_height;
  GtkRequisition sizerequisition;
  gtk_widget_size_request (notebook_label, &sizerequisition);
  default_label_height = sizerequisition.height;  
  
  button_close = gtk_button_new ();
  gtk_widget_show (button_close);
  gtk_box_pack_start (GTK_BOX (hbox9), button_close, FALSE, FALSE, 0);
  gtk_button_set_focus_on_click (GTK_BUTTON (button_close), false);

  GtkWidget *alignment2 = gtk_alignment_new (0.5, 0.5, 0, 0);
  gtk_widget_show (alignment2);
  gtk_container_add (GTK_CONTAINER (button_close), alignment2);

  GtkWidget *hbox10 = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox10);
  gtk_container_add (GTK_CONTAINER (alignment2), hbox10);

  GtkWidget *image19452 = gtk_image_new_from_stock ("gtk-close", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image19452);
  gtk_box_pack_start (GTK_BOX (hbox10), image19452, FALSE, FALSE, 0);
  gtk_widget_set_size_request (image19452, default_label_height / 3, default_label_height / 3);

  // Build the vertical paned widget. 
  // This widget separates the text and the notes.
  vpaned = gtk_vpaned_new ();
  gtk_widget_show (vpaned);
  gtk_box_pack_start (GTK_BOX (main_box), vpaned, true, true, 0);

  // Connect a signal to the size allocation, because we need it at times.
  // When only one editor is opened, there is only one tab, and everything is fine.
  // But if there are several Editors, only one is actually shown after creation.
  // The widgets in the other tabs have not yet been allocated sizes. We need
  // the height allocated to the vpaned to calculate the size of the Bible notes
  // window. If the size is not yet known, odd results might occur.
  // By connecting to this signal we get informed when the size is allocated,
  // and are enabled to re-calculate the size of the Bible notes window.
  vpaned_previous_height = 0;
  g_signal_connect_after ((gpointer) vpaned, "size_allocate", G_CALLBACK (on_vpaned_editor_size_allocate), gpointer (this));

  // Set the position to a high value to start off with: better gui.
  gtk_paned_set_position (GTK_PANED (vpaned), 200000000);

  // Vertical box.
  vbox_editor = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (vbox_editor);
  gtk_paned_pack1 (GTK_PANED (vpaned), vbox_editor, FALSE, TRUE);

  // Label showing the project. Bold.
  // Showing this label is done later, depending on the editor count.
  // label_project = gtk_label_new ("");
  // char *markup = g_markup_printf_escaped ("<span weight=\"heavy\">%s</span>", project.c_str());
  // gtk_label_set_markup (GTK_LABEL (label_project), markup);
  // g_free (markup);  
  // gtk_box_pack_start (GTK_BOX (vbox_editor), label_project, FALSE, FALSE, 0);

  // The scrolled window for the text.
  scrolledwindow_text = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow_text);
  gtk_box_pack_start (GTK_BOX (vbox_editor), scrolledwindow_text, TRUE, TRUE, 0);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow_text), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

  // The textview for the text.  
  textview_text = gtk_text_view_new ();
  gtk_widget_show (textview_text);
  gtk_container_add (GTK_CONTAINER (scrolledwindow_text), textview_text);
  gtk_text_view_set_accepts_tab (GTK_TEXT_VIEW (textview_text), FALSE);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview_text), GTK_WRAP_WORD);

  // Initialize a couple of event ids.
  textview_text_cursor_moved_delayer_event_id = 0;
  grab_focus_event_id = 0;
  position_cursor_at_verse_event_id = 0;
  undo_redo_event_id = 0;
  save_timeout_event_id = 0;  
  
  // Set some variables and connect to a couple of signals.
  g_signal_connect_after ((gpointer) textview_text, "move_cursor", G_CALLBACK (on_textview_text_move_cursor), gpointer(this));
  g_signal_connect_after ((gpointer) textview_text, "grab_focus", G_CALLBACK (on_textview_text_grab_focus), gpointer(this));
  g_signal_connect ((gpointer) textview_text, "motion-notify-event", G_CALLBACK (on_text_motion_notify_event), gpointer(this));
  g_signal_connect ((gpointer) textview_text, "event-after", G_CALLBACK (on_text_event_after), gpointer(this));
  g_signal_connect ((gpointer) textview_text, "key-press-event", G_CALLBACK (text_key_press_event), gpointer(this));
  g_signal_connect ((gpointer) textview_text, "visibility-notify-event", G_CALLBACK (screen_visibility_notify_event), gpointer(this));
  g_signal_connect ((gpointer) textview_text, "button_press_event", G_CALLBACK (on_textview_edit_button_press_event), gpointer(this));

  // Accompanying buffer.  
  textbuffer_text = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview_text));
  g_signal_connect ((gpointer) textbuffer_text, "insert_text", G_CALLBACK (on_textbuffer_edit_insert_text), gpointer(this));
  g_signal_connect ((gpointer) textbuffer_text, "delete_range", G_CALLBACK (on_textbuffer_edit_delete_text), gpointer(this));
  
  // Tag for highlighting a line.
  // The colour names can be found in rgb.txt of the X system.
  // On Fedore Core 3 this is in /usr/X11R6/lib/X11/rgb.txt
  text_line_tag = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (textbuffer_text), "linetag");
  if (text_line_tag == NULL)
    text_line_tag = gtk_text_buffer_create_tag (textbuffer_text, "linetag", "background", "light yellow", NULL);

  // Tag for highlighting search words
  reference_tag = gtk_text_buffer_create_tag (textbuffer_text, "referencetag", "background", "khaki", NULL);
  
  // Buttons to give signals.
  new_verse_signal = gtk_button_new ();
  gtk_box_pack_start (GTK_BOX (main_box), new_verse_signal, false, false, 0);
  word_double_clicked_signal = gtk_button_new ();
  gtk_box_pack_start (GTK_BOX (main_box), word_double_clicked_signal, false, false, 0);
  
  scrolledwindow_notes = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_show (scrolledwindow_notes);
  gtk_paned_pack2 (GTK_PANED (vpaned), scrolledwindow_notes, TRUE, TRUE);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow_notes), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolledwindow_notes), GTK_SHADOW_IN);

  textview_notes = gtk_text_view_new ();
  gtk_widget_show (textview_notes);
  gtk_container_add (GTK_CONTAINER (scrolledwindow_notes), textview_notes);
  gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (textview_notes), GTK_WRAP_WORD);
  // Accept no tabs. These are not in the USFM standard anyway.
  // Tab moves to next widget.
  gtk_text_view_set_accepts_tab (GTK_TEXT_VIEW (textview_notes), FALSE);
  
  // Textbuffer for the notes.
  textbuffer_notes = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview_notes));
  
  // Connect to a bunch of signals.
  g_signal_connect_after ((gpointer) textview_notes, "move_cursor", G_CALLBACK (on_textview_footnotes_move_cursor), gpointer(this));
  g_signal_connect_after ((gpointer) textview_notes, "grab_focus", G_CALLBACK (on_textview_footnotes_grab_focus), gpointer(this));
  g_signal_connect_after ((gpointer) textview_notes, "backspace", G_CALLBACK (on_textview_footnotes_backspace), gpointer(this));
  g_signal_connect_after ((gpointer) textview_notes, "cut_clipboard", G_CALLBACK (on_textview_footnotes_cut_clipboard), gpointer(this));
  g_signal_connect_after ((gpointer) textview_notes, "delete_from_cursor", G_CALLBACK (on_textview_footnotes_delete_from_cursor), gpointer(this));
  g_signal_connect ((gpointer) textview_notes, "button_press_event", G_CALLBACK (on_textview_footnotes_button_press_event), gpointer(this));
  g_signal_connect ((gpointer) textbuffer_notes, "changed", G_CALLBACK (on_textbuffer_footnotes_changed), gpointer(this));
  g_signal_connect ((gpointer) textbuffer_notes, "insert_text", G_CALLBACK (on_textbuffer_edit_insert_text), gpointer(this));
  g_signal_connect ((gpointer) textbuffer_notes, "delete_range", G_CALLBACK (on_textbuffer_edit_delete_text), gpointer(this));

  // Tag for highlighting.
  notes_line_tag = gtk_text_tag_table_lookup (gtk_text_buffer_get_tag_table (textbuffer_notes), "linetag");
  if (notes_line_tag == NULL)
    notes_line_tag = gtk_text_buffer_create_tag (textbuffer_notes, "linetag", "background", "light yellow", NULL);

  // Undo/redo system.
  un_re_do_clear_internal ();
  undo_redo_event_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 100, GSourceFunc (un_re_do_on_timeout), gpointer (this), NULL);
  
  // Automatic saving of the file, periodically.
  save_timeout_event_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 60000, GSourceFunc (mainwindow_on_save_timeout), gpointer(this), NULL);
  
  // The display will not properly show the cursor always, because it has not
  // yet been created fully on widget creation.
  g_timeout_add (1000, GSourceFunc (on_widget_creation_timeout), gpointer(this));
  
  // Grab focus.
  focus_programmatically_being_grabbed = false;
  gtk_widget_grab_focus (textview_text);  
  on_grab_focus ();
}


Editor::~Editor ()
{
  // Destroy a couple of timeout sources.
  gw_destroy_source (textview_text_cursor_moved_delayer_event_id);
  gw_destroy_source (grab_focus_event_id);
  gw_destroy_source (position_cursor_at_verse_event_id);
  gw_destroy_source (undo_redo_event_id);
  gw_destroy_source (save_timeout_event_id);
  // Destroy topmost widget. 
  // Normally, when shutting down Bibledit, Gtk will destroy this widget. 
  // But if we don't shut down Bibledit, but only remove the editor, 
  // then the widget need to be destroyed manually, 
  // which is what is done here.
  gtk_widget_destroy (main_box);
}


void Editor::book_set (unsigned int book_in)
{
  book = book_in;
}


void Editor::chapter_load (unsigned int number, vector<ustring> * lines, bool clear_undo_stack)
// Loads chapter number in the editor if lines is NULL, 
// else loads from lines.
{
  // Save chapter number, but only in case that we don't load text from "lines"
  // If we load from "lines", then the chapter number that is there already will
  // be preserved. Loss of information has been observed in cases this was
  // not followed, e.g. every chapter in which an Undo operation was done, 
  // was written to chapter 0.
  if (!lines) chapter = number;
  
  // Reset id related variable. A text tag in the editor gets an id number
  // larger than 0, and that indicates it has an associated data container.
  // The id number we take for that is the current size of the editor objects.
  // As 0 is not acceptable, fill the container with a dummy object to fill 
  // place 0.
  editorobjects.clear();
  EditorObject zero_object (eotNone);
  editorobjects.push_back (zero_object);

  // Reset note id last focused.
  previous_focused_note_id = 0;
  
  // Load text.
  bool chapter_exists = true;
  if (lines) {
    load_text (* lines);
  } else {
    vector <ustring> mylines = project_retrieve_chapter (project, book, number);
    // If this chapter does not exist, give information in the screen.
    if (mylines.empty()) {
      mylines.push_back ("This reference does not exist in the project.");
      chapter_exists = false;
    }
    load_text (mylines);
  }
 
  // Format screen.
  gtk_text_buffer_set_text (textbuffer_notes, "", -1);
  format ();
  
  // Set the editors ummodified.
  gtk_text_buffer_set_modified (textbuffer_text, false);
  gtk_text_buffer_set_modified (textbuffer_notes, false);
  
  // Set whether the buffers are editable.
  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  bool editable = projectconfig->editable_get();
  if (!chapter_exists) editable = false;
  gtk_text_view_set_editable (GTK_TEXT_VIEW (textview_text), editable);
  gtk_text_view_set_editable (GTK_TEXT_VIEW (textview_notes), editable);
  
  // Clear undo/redo stack
  if (clear_undo_stack) un_re_do_clear ();
}


void Editor::load_text (vector<ustring>& lines)
{
  // Clear buffer.
  gtk_text_buffer_set_text (textbuffer_text, "", 0);
  
  // Remove all tags from the buffer.
  GtkTextIter startiter;
  GtkTextIter enditer;
  gtk_text_buffer_get_start_iter (textbuffer_text, &startiter);
  gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
  gtk_text_buffer_remove_all_tags (textbuffer_text, &startiter, &enditer);

  // Prepare for tagging the versenumbers.
  Highlight highlight (textbuffer_text, textview_text);

  // Set the new text.
  CategorizeChapterVerse ccv (lines);
  for (unsigned int i = 0; i < ccv.verse.size (); i++) {
    // Remove the \c line.
    if (ccv.line[i].substr (0, 3) == "\\c ") {
      ParseLine parseline (ccv.line[i]);
      parseline.lines.erase (parseline.lines.begin());
      if (parseline.lines.empty()) continue;
      ccv.line[i].clear();
      for (unsigned int i2 = 0; i2 < parseline.lines.size(); i2++) {
        if (!ccv.line[i].empty()) ccv.line[i].append ("\n");
        ccv.line[i].append (parseline.lines[i2]);
      }
    }
    // Insert the tag with a name of the current verse number.
    // E.g. for verse 1 the name is "1", and for verse 1-3a it is "1-3a".
    GtkTextTag * tag = highlight.get_tag (ccv.verse[i]);
    // Insert the text tagged with this tag, at the end of the buffer.
    gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
    gtk_text_buffer_insert_with_tags (textbuffer_text, &enditer, ccv.line[i].c_str(), -1, tag, NULL);
    gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
    gtk_text_buffer_insert_with_tags (textbuffer_text, &enditer, "\n", 1, tag, NULL);
  }
}


void Editor::format_notes ()
/*
Formats the bible notes: footnotes, endnotes, crossreferences.
This function may be called with an empty notes buffer, but also with a full one.
Therefore we need to set up a strategy for how to deal with this issue.
- From the notebuffer, we extract all relevant data, and temporally store it 
  with their id.
- We clear the note buffer.
- We go through the text to look for any raw usfm bible notes code, or any
  formatted, and write the appropriate text in the notes buffer, and format 
  any raw code.
*/
{
  // Configuration data.
  extern Settings * settings;
  
  {
    // Store current note text.
    map<int, ustring> current_notes = text_note_get_all ();
    gtk_text_buffer_set_text (textbuffer_notes, "", 0);
    // Iterate through the whole buffer, 
    // reinsert the original usfm code for any note caller,
    // remove that note caller,
    // and clear the editor object after that.
    GtkTextIter iter1, iter2;
    gtk_text_buffer_get_start_iter (textbuffer_text, &iter2);
    iter1 = iter2;
    int id_last_deleted = -1;
    while (gtk_text_iter_forward_char (&iter2)) {
      // Go through the tags that apply to this iterator.
      GSList *tags = NULL, *tagp = NULL;
      tags = gtk_text_iter_get_tags (&iter1);
      for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
        GtkTextTag *tag = (GtkTextTag *) tagp->data;
        gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
        EditorObjectType objecttype = editorobjects[id].type;
        if ((objecttype == eotFootnote) || (objecttype == eotEndnote) || (objecttype == eotCrossreference) || (id == id_last_deleted)) {
          // Replace note caller with actual usfm code.
          gtk_text_buffer_delete (textbuffer_text, &iter1, &iter2);
          int offset = gtk_text_iter_get_offset (&iter2);
          gtk_text_buffer_insert (textbuffer_text, &iter1, current_notes[id].c_str(), -1);
          gtk_text_buffer_get_iter_at_offset (textbuffer_text, &iter2, offset);
          // Clear relevant objects.
          current_notes[id].clear();
          editorobjects[id].clear();
          // Store this id so that any next character with this id gets deleted too.
          // This is needed in case of note callers that are more than one character long.
          id_last_deleted = id;      
        }
      }
      // Free tags list if tags were found.
      if (tags) {
        g_slist_free (tags);
      } 
      // Next iteration.
      iter1 = iter2;
    }
  }

  // Usfm data.
  ProjectConfiguration * projectconfig = settings->projectconfig (settings->genconfig.project_get());
  Usfm usfm (projectconfig->stylesheet_get());

  // Prepare for the notes.
  EditorNote editornote (usfm);

  // Go through each line of the textbuffer.
  int number_of_lines = gtk_text_buffer_get_line_count (textbuffer_text);
  GtkTextIter iterator;
  GtkTextIter endofline;
  // Iterate through all lines.
  for (int i = 0; i < number_of_lines; i++) {
    gtk_text_buffer_get_iter_at_line (textbuffer_text, &iterator, i);
    // Ensure that also the last line, without a newline character, gets taken.
    if (i + 1 == number_of_lines) {
      gtk_text_buffer_get_end_iter (textbuffer_text, &endofline);
    } else {
      gtk_text_buffer_get_iter_at_line (textbuffer_text, &endofline, i + 1);
      gtk_text_iter_backward_char (&endofline);
    }
    // Get the line.
    ustring line = gtk_text_buffer_get_text (textbuffer_text, &iterator, &endofline, false);
    // Handle the notes: footnote, endnote, xref.
    editornote.extract (line);
    for (unsigned int i2 = 0; i2 < editornote.texts.size(); i2++) {
      // The caller, make it smaller and superscript.
      // Attach some data to the tag, to make it recognizable.
      // The rendered caller should not be editable, and this can be done using the 
      // "editable" property, but the problem is that when there are two or more
      // callers in sequence, the whole area becomes non-editable, so nothing
      // can be inserted in between. Also cutting the caller no longer works.
      // Handle this in the keyboard handlers.
      // In Paratext, when there is a footnote caller like "$$", one can insert 
      // a space in between, and after save, it only has the caller $ left in 
      // the text.
      int id = editorobjects.size();
      {
        GtkTextTag *tag;
        tag = gtk_text_buffer_create_tag (textbuffer_text, NULL,
              "editable", false, "rise", 5 * PANGO_SCALE, "scale", PANGO_SCALE_SMALL, NULL);
        g_object_set_data (G_OBJECT (tag), "id", GINT_TO_POINTER (id));
        GtkTextIter startiter, enditer;
        gtk_text_buffer_get_iter_at_line_offset (textbuffer_text, &startiter, i, editornote.positions[i2]);
        gtk_text_buffer_get_iter_at_line_offset (textbuffer_text, &enditer, i, editornote.positions[i2] + editornote.lengths[i2]);
        gtk_text_buffer_delete (textbuffer_text, &startiter, &enditer);
        gtk_text_buffer_insert_with_tags (textbuffer_text, &startiter, editornote.rendered_callers[i2].c_str(), editornote.rendered_callers[i2].length(), tag, NULL);
        EditorObject editorobject (editornote.types[i2]);
        editorobject.note_set (editornote.openers[i2], editornote.callers[i2], editornote.closers[i2]);
        editorobjects.push_back (editorobject);
      }
      // Set the note text in the textbuffer for display.
      {
        // Note caller.
        GtkTextTag *callertag;
        callertag = gtk_text_buffer_create_tag (textbuffer_notes, NULL, 
          "editable", false, "rise", 5 * PANGO_SCALE, "scale", PANGO_SCALE_SMALL, NULL);
        g_object_set_data (G_OBJECT (callertag), "id", GINT_TO_POINTER (id));
        GtkTextIter iter;
        gtk_text_buffer_get_end_iter (textbuffer_notes, &iter);
        ustring callertext = editornote.rendered_callers[i2] + " ";
        gtk_text_buffer_insert_with_tags (textbuffer_notes, &iter, callertext.c_str(), -1, callertag, NULL);
        // Note body.
        GtkTextTag *bodytag;
        bodytag = gtk_text_buffer_create_tag (textbuffer_notes, NULL, NULL);
        g_object_set_data (G_OBJECT (bodytag), "id", GINT_TO_POINTER (id));
        gtk_text_buffer_get_end_iter (textbuffer_notes, &iter);
        ustring bodytext = editornote.texts[i2];
        gtk_text_buffer_insert_with_tags (textbuffer_notes, &iter, bodytext.c_str(), -1, bodytag, NULL);
        // Add a non-editable space, with the note's id. This enables us to 
        // add text to the end of the note, which will still be considered part
        // of the note.
        GtkTextTag *spacetag;
        spacetag = gtk_text_buffer_create_tag (textbuffer_notes, NULL, "editable", false, NULL);
        g_object_set_data (G_OBJECT (spacetag), "id", GINT_TO_POINTER (id));
        gtk_text_buffer_get_end_iter (textbuffer_notes, &iter);
        gtk_text_buffer_insert_with_tags (textbuffer_notes, &iter, " ", -1, spacetag, NULL);
        // Extra info
        GtkTextTag *infotag;
        infotag = gtk_text_buffer_create_tag (textbuffer_notes, NULL, 
          "editable", false, "scale", PANGO_SCALE_SMALL, "foreground", "gray", NULL);
        gtk_text_buffer_get_end_iter (textbuffer_notes, &iter);
        ustring infotext = end_note_signature (editornote.types[i2]);
        NumberingType numberingtype;
        ustring dummy;
        biblenotes_get_note_numbering (editornote.callers[i2], numberingtype, dummy);
        infotext.append (" ");
        infotext.append (biblenotes_note_caller_human_readable (numberingtype));
        infotext.append ("\n");
        gtk_text_buffer_insert_with_tags (textbuffer_notes, &iter, infotext.c_str(), -1, infotag, NULL);
      }
    }
  }  
  
  // Set the size of the notes window depending on whether there are notes.
  bible_notes_window_set_size();
  
  // Process pending events.
  while (gtk_events_pending()) gtk_main_iteration ();
}


void Editor::chapter_save ()
{
  // If text wasn't changed, bail out.
  bool buffer_modified = gtk_text_buffer_get_modified (textbuffer_text);
  if (gtk_text_buffer_get_modified (textbuffer_notes))
    buffer_modified = true;
  if (!buffer_modified) 
    return;
  // Get chapter text.   
  GtkTextIter startiter, enditer;
  gtk_text_buffer_get_start_iter (textbuffer_text, &startiter);
  gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
  ustring chaptertext = text_get (&startiter, &enditer);
  // Clean it up a bit and divide it into lines.
  ParseLine parseline (trim (chaptertext));
  // Insert the chapter that was removed before.
  parseline.lines.insert (parseline.lines.begin(), "\\c " + convert_to_string (chapter));
  // Add verse information.
  CategorizeChapterVerse ccv (parseline.lines);
  // We have noticed people editing Bibledit's data directly. If this was done
  // with OpenOffice, and then saving it as text again, three special bytes
  // were added right at the beginning of the file, making the \c x marker
  // not being parsed correctly. We work around this possible bug this way:
  // Just update the categorized chapter information with the chapter we know
  // it should be. Give a warning if a mismatch is found.
  bool chapter_mismatch = false;
  for (unsigned int i = 0; i < ccv.chapter.size(); i++) {
    if (ccv.chapter[i] != chapter) {
      chapter_mismatch = true;
      ccv.chapter[i] = chapter;
    }
  }
  if (chapter_mismatch) {
    gw_critical ("The chapter in the editor does not match the one in the text");
  }
  // Store chapter in database.
  project_store_chapter (project, book, ccv);
  // Set the buffer to unmodified.
  gtk_text_buffer_set_modified (textbuffer_text, false);
}


void Editor::footnote_body_scroll_to ()
// Place the cursor at the beginning of the nearest footnote, 
// and scroll the widget to make the note visible.
{
  // Go through the whole editor and look up the id of the nearest footnote.
  GtkTextIter iter;
  // Calculate cursor position offset.
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
  int cursor_offset = gtk_text_iter_get_offset (&iter);
  // Go through the whole buffer and gather offsets of footnotes relative to cursor position.
  // Take smallest.
  int nearest_id = -1;
  int smallest_offset = 1000000;  
  gtk_text_buffer_get_start_iter (textbuffer_text, &iter);
  while (gtk_text_iter_forward_char (&iter)) {
    // Go through the tags that apply to this iterator.
    GSList *tags = NULL, *tagp = NULL;
    tags = gtk_text_iter_get_toggled_tags (&iter, true);
    for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
      GtkTextTag *tag = (GtkTextTag *) tagp->data;
      gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
      EditorObjectType objecttype = editorobjects[id].type;
      if ((objecttype == eotFootnote) || (objecttype == eotEndnote) || (objecttype == eotCrossreference)) {
        int offset = abs (gtk_text_iter_get_offset (&iter) - cursor_offset);
        if (offset < smallest_offset) {
          nearest_id = id;
          smallest_offset = offset;
        }
      }
    }
    // Free tags list if tags were found.
    if (tags) {
      g_slist_free (tags);
    } 
  }
  
  // Remove previous highlight for this note, if any.
  GtkTextIter highlightstartiter;
  GtkTextIter highlightenditer;
  gtk_text_buffer_get_start_iter (textbuffer_notes, &highlightstartiter);
  gtk_text_buffer_get_end_iter (textbuffer_notes, &highlightenditer);
  gtk_text_buffer_remove_tag (textbuffer_notes, notes_line_tag, &highlightstartiter, &highlightenditer);

  // Prepare for highlighting the active note.
  gtk_text_buffer_get_start_iter (textbuffer_notes, &highlightenditer);
  bool highlight_start_iter_initialized = false;
  
  // If applicable, scroll to footnote with given id and place cursor there.
  if (nearest_id > 0) {
    GtkTextTag * tag_scrolled_to = NULL;
    GtkTextIter iter;
    gtk_text_buffer_get_start_iter (textbuffer_notes, &iter);
    while (gtk_text_iter_forward_char (&iter)) {
      // Go through the tags that apply to this iterator.
      GSList *tags = NULL, *tagp = NULL;
      tags = gtk_text_iter_get_tags (&iter);
      for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
        GtkTextTag *tag = (GtkTextTag *) tagp->data;
        gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
        if (id == nearest_id) {
          if (tag != tag_scrolled_to) {
            screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_notes), &iter);
            gtk_text_buffer_place_cursor (textbuffer_notes, &iter);
            tag_scrolled_to = tag;
          }
          if (!highlight_start_iter_initialized) {
            highlightstartiter = iter;
            highlight_start_iter_initialized = true;
          }
          highlightenditer = iter;
        }
      }
      // Free tags list if tags were found.
      if (tags) {
        g_slist_free (tags);
      } 
    }
  }

  // Highlight the note.
  gtk_text_iter_forward_char (&highlightenditer);
  gtk_text_buffer_apply_tag (textbuffer_notes, notes_line_tag, &highlightstartiter, &highlightenditer);
}


bool Editor::footnote_caller_remove (bool backspace)
// Remove possible footnote caller we are at, and the associated footnote body.
// If BackSpace was pressed, remove a caller before the insertion point.
// This function is called by the keyboard handler, and returns true if a footnote
// has been deleted, so that the keyboard handler knows whether to process
// keyboard events further.
{
  // Return value.
  bool footnote_deleted = false;
  // Variables pointing to data to be deleted.
  GtkTextTag * tag_to_be_deleted = NULL;
  // Get iterator at the insertion point. If backspace go one back.
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
  if (backspace)
    gtk_text_iter_backward_char (&iter);
  // Get tags at that iterator and see whether there are any Bible notes.
  GSList *tags = NULL, *tagp = NULL;
  tags = gtk_text_iter_get_tags (&iter);
  for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
    GtkTextTag *tag = (GtkTextTag *) tagp->data;
    gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
    EditorObjectType objecttype = editorobjects[id].type;
    if ((objecttype == eotFootnote) || (objecttype == eotEndnote) || (objecttype == eotCrossreference)) {
      tag_to_be_deleted = tag;
    }
  }
  // Free tags list if tags were found.
  if (tags) {
    g_slist_free (tags);
  } 
  // If we've got a GtkTextTag to be deleted, delete the note caller.
  // As the footnote caller can be either one character, in most cases, but also
  // two like when we've numbers like 10, 11, etc., or even three in extreme
  // cases, do it so: Iterate 3 characters back, and from there look 6 ahead
  // and delete anything in between which has this tag.
  if (tag_to_be_deleted) {
    GtkTextIter startiter = iter;
    GtkTextIter enditer = iter;
    gtk_text_iter_backward_chars (&iter, 3);
    for (unsigned int i = 0; i < 6; i++) {
      if (gtk_text_iter_begins_tag (&iter, tag_to_be_deleted)) {
        startiter = iter;
      }
      if (gtk_text_iter_ends_tag (&iter, tag_to_be_deleted)) {
        enditer = iter;
      }
      gtk_text_iter_forward_char (&iter);
    }
    gtk_text_buffer_delete (textbuffer_text, &startiter, &enditer);    
    footnote_deleted = true;
  }
  // If something was deleted, reformat the screen.
  if (footnote_deleted)
    format ();
  // Returnvalue shows whether we deleted a note.
  return footnote_deleted;
}


void Editor::footnote_remove_empty_ones ()
{
  GtkTextIter startiter, enditer;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &startiter, gtk_text_buffer_get_insert (textbuffer_text));
  enditer = startiter;
  gtk_text_iter_forward_char (&enditer);
  gtk_text_iter_order (&startiter, &enditer);
  if (text_get (&startiter, &enditer).empty()) {
    footnote_caller_remove (false);
  }
}


ustring Editor::text_get (GtkTextIter * startiter, GtkTextIter * enditer)
// Iterate through the buffer and collect and assemble usfm text.
{
  // Some variables needed.    
  ustring text;
  GtkTextIter iter = * startiter;
  // Go through this part of the buffer.  
  while (gtk_text_iter_in_range (&iter, startiter, enditer)) {
    // Go through the tags that apply to this iterator.
    GSList *tags = NULL, *tagp = NULL;
    tags = gtk_text_iter_get_tags (&iter);
    bool tagsfound = false;
    for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
      GtkTextTag *tag = (GtkTextTag *) tagp->data;
      gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
      if (id != 0) {
        tagsfound = true;
        EditorObjectType objecttype = editorobjects[id].type;
        switch (objecttype) {
          case eotNone: 
            break;
          case eotFootnote:
          case eotEndnote:
          case eotCrossreference:
          {
            // In case of a tag for a note, do not output the normal text, but
            // at the start of the tag, output the stored data for this note.
            if (gtk_text_iter_begins_tag (&iter, tag)) {
              text.append (text_note_get (id));
            }
            break;
          }
          default :
          {
          }
        }
      }
    }
    // Free tags list if tags were found.
    if (tags) {
      g_slist_free (tags);
    } 
    if (!tagsfound) {
      // If no tags found, handle normal text.
      GtkTextIter iter2 = iter;
      gtk_text_iter_forward_char (&iter2);
      text.append (gtk_text_iter_get_text (&iter, &iter2));
    }
    // Next iteration.
    gtk_text_iter_forward_char (&iter);
  }
  // Give result.
  return text;
}


ustring Editor::text_get_selection ()
// Convenience function. Returns selected usfm code.
{
  GtkTextIter startiter, enditer;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &startiter, gtk_text_buffer_get_insert (textbuffer_text));
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &enditer, gtk_text_buffer_get_selection_bound (textbuffer_text));
  gtk_text_iter_order (&startiter, &enditer);
  return text_get (&startiter, &enditer);
}


ustring Editor::text_get_all ()
// Convenience function. Returns all usfm code.
{
  GtkTextIter startiter, enditer;
  gtk_text_buffer_get_start_iter (textbuffer_text, &startiter);
  gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
  gtk_text_iter_order (&startiter, &enditer);
  return text_get (&startiter, &enditer);
}


ustring Editor::text_note_get (gint id)
// Gets the usfm text for a note with id = n.
// It gets the data from the objects stored in memory and from the note textview.
{
  // Variables.
  ustring text, opener, caller, closer;
  // The stored object.
  EditorObject editorobject (editorobjects[id]);
  editorobject.note_get (opener, caller, closer);
  // Start assembling usfm text.
  text.append (opener);
  text.append (caller);
  text.append (" ");
  // Extract the text from the notes textview.
  // There is a problem that text, added at the very end of a note, does not
  // have the same tag applied as the existing note text. Solve that by 
  // iterating beyond the note, till a certain signature is encountered.
  ustring notetext;
  bool was_right_id = false;
  GtkTextIter iter1, iter2;
  gtk_text_buffer_get_start_iter (textbuffer_notes, &iter2);
  iter1 = iter2;
  while (gtk_text_iter_forward_char (&iter2)) {
    // Go through the tags that apply to this iterator.
    GSList *tags = NULL, *tagp = NULL;
    tags = gtk_text_iter_get_tags (&iter1);
    for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
      GtkTextTag *tag = (GtkTextTag *) tagp->data;
      gint id2 = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
      if (id2 == id) {
        was_right_id = true;
      } 
    }
    if (was_right_id) {
      notetext.append (gtk_text_iter_get_text (&iter1, &iter2));
    }
    // Free tags list if tags were found.
    if (tags) {
      g_slist_free (tags);
    } 
    // Next iteration.
    iter1 = iter2;
  }
  // Change \n to a space.
  replace_text (notetext, "\n", " ");
  // Clear out first bit: caller and space.
  notetext.erase (0, notetext.find (" "));
  notetext = trim (notetext);
  // Clear out anything after signature.
  {
    ustring signature = trim (end_note_signature (editorobject.type));
    size_t position = notetext.find (signature);
    if (position != string::npos) {
      notetext.erase (position, 1000000);
    }
    notetext = trim (notetext);
  }
  text.append (notetext);
  text.append (closer);
  // If there is no actual content, there is no note at all.
  if (notetext.empty())
    text.clear();
  // Return result.
  return text;
}


map <int, ustring> Editor::text_note_get_all ()
// Gets a map with all note text in it.
{
  // Result variable.
  map <int, ustring> result;
  
  // Get all available ids in the notes buffer.
  vector<int> ids;
  {
    set<int> map_ids;
    GtkTextIter iter1, iter2;
    gtk_text_buffer_get_start_iter (textbuffer_notes, &iter2);
    iter1 = iter2;
    while (gtk_text_iter_forward_char (&iter2)) {
      // Go through the tags that apply to this iterator.
      GSList *tags = NULL, *tagp = NULL;
      tags = gtk_text_iter_get_tags (&iter1);
      for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
        GtkTextTag *tag = (GtkTextTag *) tagp->data;
        gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
        if (id > 0) {
          EditorObjectType objecttype = editorobjects[id].type;
          if ((objecttype == eotFootnote) || (objecttype == eotEndnote) || (objecttype == eotCrossreference)) {        
            map_ids.insert (id);
          }
        }
      }
      // Free tags list if tags were found.
      if (tags) {
        g_slist_free (tags);
      } 
      // Next iteration.
      iter1 = iter2;
    }
    ids.assign (map_ids.begin(), map_ids.end());
  }

  // Get text for all ids.
  for (unsigned int i = 0; i < ids.size(); i++) {
    result[ids[i]] = text_note_get (ids[i]);
  }
  
  // Result.
  return result;
}


void Editor::format ()
{
  format_notes ();
  footnote_body_scroll_to ();
}


void Editor::footnote_caller_scroll_to ()
/*
This function is called whenever there is cursor movement in the footnote editor.
It scrolls the text editor to the appropriate footnote caller.
*/
{
  // Get the id of the footnote that has the cursor.
  int footnote_id = note_id_at_cursor_get ();
  // If we did not find any relevant id, bail out.
  if (footnote_id == 0)
    return;
  // Go through the whole editor and look for this id. Position cursor there.  
  GtkTextIter iter;
  gtk_text_buffer_get_start_iter (textbuffer_text, &iter);
  while (gtk_text_iter_forward_char (&iter)) {
    // Go through the tags that start at this iterator.
    GSList *tags = NULL, *tagp = NULL;
    tags = gtk_text_iter_get_toggled_tags (&iter, true);
    for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
      GtkTextTag *tag = (GtkTextTag *) tagp->data;
      gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
      if (id == footnote_id) {
        // Place cursor and scroll.
        screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_text), &iter);
        gtk_text_buffer_place_cursor (textbuffer_text, &iter);
        // Bail out, but first free memory.
        if (tags)
          g_slist_free (tags);
        return;
      }
    }
    // Free tags list if tags were found.
    if (tags) {
      g_slist_free (tags);
    } 
  }
}


ustring Editor::end_note_signature (EditorObjectType objecttype)
{
  ustring signature = "(";
  if (objecttype == eotFootnote) signature.append ("footnote");
  if (objecttype == eotEndnote) signature.append ("endnote");
  if (objecttype == eotCrossreference) signature.append ("crossreference");
  signature.append (")");
  return signature;  
}


void Editor::show_quick_references (GtkWidget * textview, const ustring& project, unsigned int book, bool reread)
// Takes the text of the references in the note that has the cursor,
// and shows that text in the quick reference area.
{
  // Get the textbuffer.
  GtkTextBuffer * buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  // Get the id of the note that has the cursor.
  int note_id = note_id_at_cursor_get ();
  // If we did not find any relevant id, clear text, and bail out.
  if (note_id == 0) {
    gtk_text_buffer_set_text (buffer, "", 0);
    previous_focused_note_id = note_id;
    return;
  }
  // If rereading the note, clear variable.
  if (reread) previous_focused_note_id = 0;
  // If we already have shown this note, bail out.
  if (note_id == previous_focused_note_id) return;
  previous_focused_note_id = note_id;
  // Get the text of the note.
  ustring notetext = text_note_get (note_id);
  // Get the language of the project.
  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  ustring language = projectconfig->language_get();
  // Extract references.
  vector<Reference> references = extract_abbreviated_references (language, book, notetext);
  // Set the text in the quick reference area.
  ustring quickreferences;
  for (unsigned int i = 0; i < references.size(); i++) {
    quickreferences.append (references[i].human_readable (language));
    quickreferences.append (" ");
    ustring text = project_retrieve_verse (project, references[i].book, references[i].chapter, references[i].verse);
    if (text.empty()) {
      quickreferences.append ("<empty>");
    } else {
      CategorizeLine cl (text);
      quickreferences.append (cl.verse);
    }
    quickreferences.append ("\n");
  }
  gtk_text_buffer_set_text (buffer, quickreferences.c_str(), -1);
}


gint Editor::note_id_at_cursor_get ()
// Get the id of the note that has the cursor.
{
  int note_id = 0;
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_mark (textbuffer_notes, &iter, gtk_text_buffer_get_insert (textbuffer_notes));
  GSList *tags = NULL, *tagp = NULL;
  tags = gtk_text_iter_get_tags (&iter);
  for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
    GtkTextTag *tag = (GtkTextTag *) tagp->data;
    gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
    if (id > 0)
      note_id = id;
  }
  if (tags) {
    g_slist_free (tags);
  } 
  return note_id;
}


void Editor::on_textview_text_move_cursor (GtkTextView * textview, GtkMovementStep step, gint count,
                                           gboolean extend_selection, gpointer user_data)
{
  ((Editor *) user_data)->on_textview_text_cursor_moved_delayer ();
}


void Editor::on_textview_text_cursor_moved_delayer ()
{
  // Keep postponing the actual handler if a new cursor movement was detected
  // before the previous one was processed: destroy the GSource.
  gw_destroy_source (textview_text_cursor_moved_delayer_event_id);
  textview_text_cursor_moved_delayer_event_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 100, GSourceFunc (on_textview_text_cursor_moved_delayer_handler), gpointer(this), NULL);
}


bool Editor::on_textview_text_cursor_moved_delayer_handler (gpointer user_data)
{
  ((Editor *) user_data)->on_textview_text_cursor_moved ();
  return false;
}


void Editor::on_textview_text_cursor_moved ()
/*
When the cursor has moved, the navigation system needs to be updated
so that it shows the right reference. If the user was, for example
on Matthew 1:10, and the cursor moves, the move might have brought him
to another reference, though this is not necessarily so. Therefore, as we 
don't know where the user is now after the cursor moved, we need to find
it out. The book is known, the chapter is known, because both stay the same.
The only thing we don't know is the verse. 
*/
{
  static int previous_cursor_line = 0;
  // Initializing the value to 0 instead of -1 solves a Segmentation fault 
  // exception when no document was loaded.
  GtkTextIter iterator;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iterator, gtk_text_buffer_get_insert (textbuffer_text));
  int line_of_cursor = gtk_text_iter_get_line (&iterator);
  if (line_of_cursor != previous_cursor_line) {
    previous_cursor_line = line_of_cursor;
    Highlight highlight (textbuffer_text, textview_text);
    highlight.line_at_cursor (text_line_tag);
    // Emit the signal that the verse has changed.
    gtk_button_clicked (GTK_BUTTON (new_verse_signal));
  }
  // If there is any note, make that note visible in the textview.
  footnote_body_scroll_to ();
}


ustring Editor::verse_number_get ()
/*
This gets the verse number the cursor is on.
At the time the text was loaded in the editor, we added named tags.
We retrieve these tags here and decide from the name they have, which verse
we are on.
*/
{
  // Variable for the verse.
  ustring verse = "0";
  // Get an iterator at the cursor location.
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
  // There is a special case, and that is when we move the cursor to the very 
  // end of the textbuffer. In this case it won't give the right verse.
  // Handle this case here: Move the iterator one character back.
  GtkTextIter enditer;
  gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
  if (gtk_text_iter_compare (&iter, &enditer) == 0) {
    gtk_text_iter_backward_char (&iter);
  }
  // Get all tags at this iterator and look for the named tag that has a property
  // named "verse-tag" with a value of "1". If found, the name of this tag is 
  // the verse number we are now on.
  GSList *tags = NULL, *tagp = NULL;
  tags = gtk_text_iter_get_tags (&iter);
  for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
    GtkTextTag *tag = (GtkTextTag *) tagp->data;
    gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "verse-tag"));
    if (id == 1) {
      gchar *strval;
      g_object_get (tag, "name", &strval, NULL);
      verse = strval;
      if (strval)
        g_free (strval);
    }
  }
  // Return the verse info found.
  return verse; 
}



void Editor::on_textview_text_grab_focus (GtkWidget * widget, gpointer user_data)
{
  ((Editor *) user_data)->on_grab_focus ();
}


void Editor::on_grab_focus ()
{
  // The purpose of this is to adjust the reference to the new position,
  // wherever the user clicks in the text.
  if (focus_programmatically_being_grabbed)
    return;
  gw_destroy_source (grab_focus_event_id);
  grab_focus_event_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 10, GSourceFunc (on_grab_focus_delayer_timeout), gpointer(this), NULL);
}


bool Editor::on_grab_focus_delayer_timeout (gpointer data)
{
  return ((Editor *) data)->on_grab_focus_delayed_handler ();
}


bool Editor::on_grab_focus_delayed_handler ()
/*
If the user clicks in the editor window, and straight after that
the verse where the cursor is, is asked, we get a reply 
indicating the verse where the cursor was prior to clicking.
This delayed handler solves that.
*/
{
  // Emit the signal that the verse has changed.
  gtk_button_clicked (GTK_BUTTON (new_verse_signal));
  // Highlight the possible new line.
  Highlight highlight (textbuffer_text, textview_text);
  highlight.line_at_cursor (text_line_tag);
  // If there is any note, make that note visible in the textview.
  footnote_body_scroll_to ();
  // Ready.
  return false;
}


bool Editor::un_re_do_on_timeout (gpointer data)
{
  ((Editor *) data)->un_re_do_timeout ();
  return true;
}


void Editor::un_re_do_timeout ()
{
  if (un_re_do_tick ()) {
    un_re_do_store_snapshot ();
  }
}


// Value that defines how soon after a change in the buffer a snapshot is taken.
// Setting it too soon gives many snapshots for little changes.
// Setting it too late does not give enough snapshots.
#define UNDO_TIME 7


void Editor::un_re_do_clear ()
{
  un_re_do_clear_internal ();
  un_re_do_previous_snapshot = text_get_all ();
  un_re_do_store_cursorpositions ();
}


void Editor::un_re_do_clear_internal ()
{
  // Clear the whole undo stack.
  un_re_do_cursorpositions_text.clear ();
  un_re_do_cursorpositions_notes.clear ();
  un_re_do_snapshots.clear ();
  un_re_do_pointer = 0; // This is normally equal to the size of the stack.
  un_re_do_timer = UNDO_TIME;
}


bool Editor::un_re_do_tick ()
/*
  Is called at regular intervals from the main window's timer.
  If it is time to store an image of the buffer, that is, if a change needs
  to be recorded, it returns true, else false.
*/
{
  un_re_do_timer++;
  if (un_re_do_timer == UNDO_TIME) {
    return true;
  }
  return false;
}


void Editor::un_re_do_store_snapshot ()
{
  // If the pointer does not point to the end of the buffer, erase from there to the end.
  if (un_re_do_pointer != un_re_do_snapshots.size ()) {
    int n = un_re_do_snapshots.size () - un_re_do_pointer;
    for (int i = 0; i < n; i++) {
      un_re_do_cursorpositions_text.pop_back ();
      un_re_do_cursorpositions_notes.pop_back ();
      un_re_do_snapshots.pop_back ();
    }
  }
  // Store the new data in the stack.
  un_re_do_cursorpositions_text.push_back (un_re_do_previous_cursorposition_text);
  un_re_do_cursorpositions_notes.push_back (un_re_do_previous_cursorposition_notes);
  un_re_do_snapshots.push_back (un_re_do_previous_snapshot);
  // Retrieve the raw usfm code which is now in the buffer and store for next time.
  un_re_do_previous_snapshot = text_get_all ();
  // Retrieve the cursor positions, store for next time.
  un_re_do_store_cursorpositions ();
  // Trim buffer it if it grows too large.
  if (un_re_do_snapshots.size () > 30) {
    un_re_do_cursorpositions_text.erase (un_re_do_cursorpositions_text.begin ());
    un_re_do_cursorpositions_notes.erase (un_re_do_cursorpositions_notes.begin ());
    un_re_do_snapshots.erase (un_re_do_snapshots.begin ());
  }
  // Update the pointer.
  un_re_do_pointer = un_re_do_snapshots.size ();
}


void Editor::un_re_do_text_changed ()
{
  un_re_do_timer = 0;
  un_re_do_store_cursorpositions ();
}


void Editor::undo ()
{
  if (can_undo ()) {
    // Update pointer.
    un_re_do_pointer--;
    // Restore text.
    ParseLine parseline (un_re_do_snapshots[un_re_do_pointer]);
    chapter_load (0, &parseline.lines, false);
    format ();
    un_re_do_cursorpositions_set (un_re_do_cursorpositions_text[un_re_do_pointer], un_re_do_cursorpositions_notes[un_re_do_pointer]);
    // Scroll to the cursor location.
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
    screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_text), &iter);
    // Let the footnote scroll to the right position.
    footnote_body_scroll_to ();
    // Update shadow data.
    un_re_do_previous_snapshot = un_re_do_snapshots[un_re_do_pointer];
    un_re_do_previous_cursorposition_text = un_re_do_cursorpositions_text[un_re_do_pointer];
    un_re_do_previous_cursorposition_notes = un_re_do_cursorpositions_notes[un_re_do_pointer];
  }
  /*
  Prevent taking a snapshots of this change again, which would lead to an 
  infinite loop. Undoing an image triggers a reformat, which triggers plenty of 
  insert and delete signals on the textbuffers, because reformatting involves 
  inserting and deleting in the buffer.
  */
  un_re_do_timer = UNDO_TIME;
}


void Editor::redo ()
{
  if (can_redo ()) {
    // Restore text.
    ParseLine parseline (un_re_do_snapshots[un_re_do_pointer]);
    chapter_load (0, &parseline.lines, false);
    format ();
    un_re_do_cursorpositions_set (un_re_do_cursorpositions_text[un_re_do_pointer], un_re_do_cursorpositions_notes[un_re_do_pointer]);
    // Scroll to the cursor location.
    GtkTextIter iter;
    gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
    screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_text), &iter);
    // Let the footnote scroll to the right position.
    footnote_body_scroll_to ();
    // Update shadow data.
    un_re_do_previous_snapshot = un_re_do_snapshots[un_re_do_pointer];
    un_re_do_previous_cursorposition_text = un_re_do_cursorpositions_text[un_re_do_pointer];
    un_re_do_previous_cursorposition_notes = un_re_do_cursorpositions_notes[un_re_do_pointer];
    // Update pointer.
    un_re_do_pointer++;
  }
  // Prevent taking a snapshots of this change.
  un_re_do_timer = UNDO_TIME;
}


bool Editor::can_undo ()
{
  return (un_re_do_pointer > 0);
}


bool Editor::can_redo ()
{
  return (un_re_do_pointer < un_re_do_snapshots.size ());
}


void Editor::un_re_do_cursorpositions_set (int textposition, int notesposition)
{
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_offset (textbuffer_text, &iter, textposition);
  gtk_text_buffer_place_cursor (textbuffer_text, &iter);
  gtk_text_buffer_get_iter_at_offset (textbuffer_notes, &iter, notesposition);
  gtk_text_buffer_place_cursor (textbuffer_notes, &iter);
}


void Editor::un_re_do_store_cursorpositions ()
{
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
  un_re_do_previous_cursorposition_text = gtk_text_iter_get_offset (&iter);
  gtk_text_buffer_get_iter_at_mark (textbuffer_notes, &iter, gtk_text_buffer_get_insert (textbuffer_notes));
  un_re_do_previous_cursorposition_notes = gtk_text_iter_get_offset (&iter);
}


void Editor::set_font ()
{
  // Set font for the text editor
  // The note editor gets the same font, but a bit smaller.
  PangoFontDescription *font_desc = NULL;
  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  if (!projectconfig->editor_font_default_get ()) {
    font_desc = pango_font_description_from_string (projectconfig->editor_font_name_get().c_str ());
  }
  gtk_widget_modify_font (textview_text, font_desc);
  if (!projectconfig->editor_font_default_get ()) {
    gint fontsize = pango_font_description_get_size (font_desc);
    fontsize = gint (fontsize * PANGO_SCALE_SMALL);
    pango_font_description_set_size (font_desc, fontsize);
  }
  gtk_widget_modify_font (textview_notes, font_desc);
  if (font_desc) pango_font_description_free (font_desc);
    
  // Set the colours for both of the editors.
  if (projectconfig->editor_default_color_get ()) {
    color_widget_default (textview_text);
    color_widget_default (textview_notes);
  } else {
    color_widget_set (textview_text,
                      projectconfig->editor_normal_text_color_get (),
                      projectconfig->editor_background_color_get (),
                      projectconfig->editor_selected_text_color_get (),
                      projectconfig->editor_selection_color_get ());
    color_widget_set (textview_notes,
                      projectconfig->editor_normal_text_color_get (),
                      projectconfig->editor_background_color_get (),
                      projectconfig->editor_selected_text_color_get (),
                      projectconfig->editor_selection_color_get ());
  }
  
  // Set predominant text direction for the editors.
  if (projectconfig->right_to_left_get ()) {
    gtk_widget_set_direction (textview_text, GTK_TEXT_DIR_RTL);
  } else {
    gtk_widget_set_direction (textview_text, GTK_TEXT_DIR_LTR);
  }  
}


void Editor::position_cursor_at_verse (const ustring& cursorposition, bool focus)
// This function starts the procedure to move the cursor of the editor to the 
// verse given.
{
  position_cursor_at_verse_cursorposition = cursorposition;
  position_cursor_at_verse_focus = focus;
  // Keep postponing the actual handler if a new positioning request was 
  // received before the previous one was processed: destroy the GSource.
  gw_destroy_source (position_cursor_at_verse_event_id);
  position_cursor_at_verse_event_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 100, GSourceFunc (position_cursor_at_verse_postponer_handler), gpointer(this), NULL);
}


bool Editor::position_cursor_at_verse_postponer_handler (gpointer user_data)
{
  ((Editor *) user_data)->position_cursor_at_verse_executer ();
  return false;
}


void Editor::position_cursor_at_verse_executer ()
{
  // Find out whether we need to reposition the cursor. We will not move the 
  // cursor or scroll to it when the cursor is already on the right verse.
  bool reposition = position_cursor_at_verse_cursorposition != verse_number_get ();
  // When the cursor is at the end of the buffer, it usually means that the buffer has just been filled.
  // So the cursor does need to be repositioned in this case.
  GtkTextIter cursor_iterator;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &cursor_iterator, gtk_text_buffer_get_insert (textbuffer_text));
  GtkTextIter enditer;
  gtk_text_buffer_get_end_iter (textbuffer_text, &enditer);
  if (gtk_text_iter_compare (&cursor_iterator, &enditer) == 0)
    reposition = true;

  // Highlighting.
  Highlight highlight (textbuffer_text, textview_text);
  
  // Do the repositioning if needed.
  if (reposition) {
    // Grab focus here to get the scrolling done properly, and the user can type 
    // in the editor. But this is only done if requested.
    if (position_cursor_at_verse_focus) {
      focus_programmatically_being_grabbed = true;
      gtk_widget_grab_focus (textview_text);
      focus_programmatically_being_grabbed = false;
    }
    if ((position_cursor_at_verse_cursorposition == "0")
      || (position_cursor_at_verse_cursorposition.empty())) {
      // Verse 0 or empty: beginning of file.
      if (gtk_text_buffer_get_line_count (textbuffer_text) > 0) {
        GtkTextIter iter;
        gtk_text_buffer_get_iter_at_line (textbuffer_text, &iter, 0);
        gtk_text_buffer_place_cursor (textbuffer_text, &iter);
        screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_text), &iter);
        while (gtk_events_pending ()) gtk_main_iteration ();
      }
    } else {
      // Go through all the lines and find out about the verse.
      for (int i = 0; i < gtk_text_buffer_get_line_count (textbuffer_text); i++) {
        GtkTextIter begin;
        GtkTextIter end;
        gtk_text_buffer_get_iter_at_line (textbuffer_text, &begin, i);
        gtk_text_buffer_get_iter_at_line (textbuffer_text, &end, i + 1);
        ustring verse;
        GSList *tags = NULL, *tagp = NULL;
        tags = gtk_text_iter_get_tags (&begin);
        for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
          GtkTextTag *tag = (GtkTextTag *) tagp->data;
          gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "verse-tag"));
          if (id == 1) {
            gchar *strval;
            g_object_get (tag, "name", &strval, NULL);
            verse = strval;
            if (strval)
              g_free (strval);
          }
        }
        // Position the cursor at the requested verse, if the verse is there.
        // Also look whether the verse is in a sequence or range of verses.
        bool position_here = (verse == position_cursor_at_verse_cursorposition);
        unsigned int verse_int = convert_to_int (position_cursor_at_verse_cursorposition);
        vector <unsigned int> combined_verses = verse_range_sequence (verse);
        for (unsigned int i2 = 0; i2 < combined_verses.size(); i2++) {
          if (verse_int == combined_verses[i2]) {
            position_here = true;
            position_cursor_at_verse_cursorposition = verse;
          }        
        }
        if (position_here) {
          // Move the cursor to it.
          // Move it to the beginning of the text, if there is any.
          int offset = 4 + verse.length ();
          ustring s = gtk_text_buffer_get_text (textbuffer_text, &begin, &end, false);
          int linelength = s.length ();
          linelength--;
          if (offset >= linelength)
            offset = linelength;
          gtk_text_buffer_get_iter_at_line_offset (textbuffer_text, &begin, i, offset);
          gtk_text_buffer_place_cursor (textbuffer_text, &begin);
          while (gtk_events_pending ()) gtk_main_iteration ();
          // Scroll also to it. It will scroll to the beginning of the text after the verse marker.
          // Alignment is needed to put the line being edited near the top of the window.
          screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_text), &begin);
          while (gtk_events_pending ()) gtk_main_iteration ();
          // Bail out.
          break;
        }
      }
    }
    // Highlight search words.
    highlight.searchwords (reference_tag, position_cursor_at_verse_cursorposition);
  }
  highlight.line_at_cursor (text_line_tag);
}


bool Editor::mainwindow_on_save_timeout (gpointer data)
{
  return ((Editor *) data)->on_save_timeout ();
}


bool Editor::on_save_timeout ()
{
  chapter_save ();
  return true;
}


gboolean Editor::on_text_motion_notify_event (GtkWidget *textview, GdkEventMotion *event, gpointer user_data)
{
  return ((Editor *) user_data)->text_motion_notify_event (textview, event);
}

  
gboolean Editor::text_motion_notify_event (GtkWidget *textview, GdkEventMotion *event)
// Update the cursor image if the pointer moved. 
{
  // Switched off as we have no need of a hand-cursor at present.
  return false;
  gint x, y;
  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (textview), GTK_TEXT_WINDOW_WIDGET,
                                         gint(event->x), gint(event->y), &x, &y);
  screen_set_cursor_hand_or_regular (GTK_TEXT_VIEW (textview), x, y);
  gdk_window_get_pointer (textview->window, NULL, NULL, NULL);
  return false;
}


void Editor::on_text_event_after (GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  ((Editor *) user_data)->text_event_after (widget, event);  
}


void Editor::text_event_after (GtkWidget *textview, GdkEvent *ev)
/* Links can also be activated by clicking.
 */
{
  GtkTextIter start, end, iter;
  GtkTextBuffer *buffer;
  GdkEventButton *event;
  gint x, y;
  if (ev->type != GDK_BUTTON_RELEASE)
    return;
  event = (GdkEventButton *)ev;
  if (event->button != 1)
    return;
  buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (textview));
  // We shouldn't follow a link if the user has selected something.
  gtk_text_buffer_get_selection_bounds (buffer, &start, &end);
  if (gtk_text_iter_get_offset (&start) != gtk_text_iter_get_offset (&end))
    return;
  gtk_text_view_window_to_buffer_coords (GTK_TEXT_VIEW (textview), GTK_TEXT_WINDOW_WIDGET,
                                         gint(event->x), gint(event->y), &x, &y);
  gtk_text_view_get_iter_at_location (GTK_TEXT_VIEW (textview), &iter, x, y);
  text_edit_if_link (textview, &iter);
}


void Editor::text_edit_if_link (GtkWidget *textview, GtkTextIter *iter)
/* Looks at all tags covering the position of iter in the text view, 
 * and if one of them is a link, follow it by showing the page identified
 * by the data attached to it.
 */
{
  GSList *tags = NULL, *tagp = NULL;
  tags = gtk_text_iter_get_tags (iter);
  for (tagp = tags;  tagp != NULL;  tagp = tagp->next) {
    GtkTextTag *tag = (GtkTextTag *) tagp->data;
    gint id = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (tag), "id"));
    if (id != 0) {
      EditorObject editorobject (editorobjects[id]);
      switch (editorobject.type) {
        case eotNone: 
          break;
        case eotFootnote:
        case eotEndnote:
        case eotCrossreference:
          break;
      }
    }
  }
  if (tags) 
    g_slist_free (tags);
}


gboolean Editor::text_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
  return ((Editor *) user_data)->on_text_key_press_event (widget, event);
}


gboolean Editor::on_text_key_press_event (GtkWidget *widget, GdkEventKey *event)
// Handle the keyboard events in the text editor.
{
  switch (event->keyval) {
    case GDK_BackSpace: 
    case GDK_Delete: 
    case GDK_KP_Delete:
    {
      if (footnote_caller_remove (event->keyval == GDK_BackSpace)) {
        return true;
      }
      break;
    }
    default:
    {
      break;
    }
  }
  return false;
}


void Editor::on_textview_footnotes_move_cursor (GtkTextView * textview, GtkMovementStep step, gint count, gboolean extend_selection, gpointer user_data)
{
  ((Editor *) user_data)->on_textview_footnotes_cursor_moved ();
}


void Editor::on_textview_footnotes_cursor_moved ()
{
  footnote_caller_scroll_to ();
  extern Settings * settings;
  show_quick_references (quick_references_textview, settings->genconfig.project_get(), settings->genconfig.book_get (), false);
}


void Editor::on_textview_footnotes_grab_focus (GtkWidget * widget, gpointer user_data)
{
  ((Editor *) user_data)->on_textview_footnotes_focus_grabbed ();
}


void Editor::on_textview_footnotes_focus_grabbed ()
{
  // When the user clicks in the textview, the cursor still seems to be at the
  // previous position. Therefore a timeout is introduced to give the cursor
  // time to reposition properly.
  g_timeout_add (1, GSourceFunc (on_textview_footnotes_focus_grabbed_delayed), gpointer (this));
}


bool Editor::on_textview_footnotes_focus_grabbed_delayed (gpointer data)
{
  ((Editor *) data)->on_textview_footnotes_focus_grabbed_timeout ();
  return false;  
}


void Editor::on_textview_footnotes_focus_grabbed_timeout ()
{
  footnote_caller_scroll_to ();
  extern Settings * settings;
  show_quick_references (quick_references_textview, settings->genconfig.project_get(), settings->genconfig.book_get (), true);
}


void Editor::on_textview_footnotes_backspace (GtkTextView *textview, gpointer user_data)
{
  ((Editor *) user_data)->on_textview_footnotes_erase ();
}


void Editor::on_textview_footnotes_cut_clipboard (GtkTextView *textview, gpointer user_data)
{
  ((Editor *) user_data)->on_textview_footnotes_erase ();
}


void Editor::on_textview_footnotes_delete_from_cursor (GtkTextView *textview, GtkDeleteType type, gint count, gpointer user_data)
{
  ((Editor *) user_data)->on_textview_footnotes_erase ();
}


void Editor::on_textview_footnotes_erase ()
// If all text has been erased, also erase the whole note from the main editor.
{
  footnote_remove_empty_ones ();
}


void Editor::on_textbuffer_footnotes_changed (GtkEditable * editable, gpointer user_data)
{
  ((Editor *) user_data)->on_textbuffer_footnotes ();
}


void Editor::on_textbuffer_footnotes ()
{
  extern Settings * settings;
  show_quick_references (quick_references_textview, settings->genconfig.project_get(), settings->genconfig.book_get (), true);
}


void Editor::quick_references_textview_set (GtkWidget * widget)
{
  quick_references_textview = widget;
}


gboolean Editor::on_textview_edit_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  ((Editor *) user_data)->on_texteditor_click (widget, event);
  return FALSE;
}


gboolean Editor::on_textview_footnotes_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  ((Editor *) user_data)->on_texteditor_click (widget, event);
  return FALSE;
}


void Editor::on_texteditor_click (GtkWidget * widget, GdkEventButton *event)
{
  // Double-clicking sends the word to Toolbox.
  if (event->type == GDK_2BUTTON_PRESS) {
    // Get the word.
    GtkTextIter enditer;
    GtkTextIter startiter;
    gtk_text_buffer_get_iter_at_mark (textbuffer_text, &enditer, gtk_text_buffer_get_insert (textbuffer_text));
    if (!gtk_text_iter_ends_word (&enditer))
      gtk_text_iter_forward_word_end (&enditer);
    startiter = enditer;
    gtk_text_iter_backward_word_start (&startiter);
    word_double_clicked_text = gtk_text_buffer_get_text (textbuffer_text, &startiter, &enditer, false);
    // Signal to have it sent to Toolbox.
    gtk_button_clicked (GTK_BUTTON (word_double_clicked_signal));
  }
}


void Editor::on_textbuffer_edit_insert_text (GtkTextBuffer * textbuffer, GtkTextIter * arg1, gchar * arg2, gint arg3, gpointer user_data)
{
  // As the documentation of this callback is not so good (gtk+ 2.6 was checked),
  // here follows some extra info.
  // textbuffer is the buffer where the text was inserted.
  // arg1 is the iterator pointing to the location where the insert was done.
  // arg2 is the string that was inserted.
  // arg3 is the length of the string that was inserted.
  ((Editor *) user_data)->on_textbuffer_edit_insert_delete ();
}


void Editor::on_textbuffer_edit_delete_text (GtkTextBuffer * textbuffer, GtkTextIter * arg1, GtkTextIter * arg2, gpointer user_data)
{
  // As the documentation of this callback is not so good (gtk+ 2.6 was checked),
  // here follows some extra info.
  // textbuffer is the buffer where text was deleted from.
  // arg1 is the iterator pointing to the beginning of the text to be deleted.
  // arg2 is the iterator pointing to the end of the text to be deleted.
  // Function is called before the actual delete operation takes place.
  ((Editor *) user_data)->on_textbuffer_edit_insert_delete ();
}


void Editor::on_textbuffer_edit_insert_delete ()
// Gets fired on inserting or deleting text in editor.
{
  un_re_do_text_changed ();
}



bool Editor::on_widget_creation_timeout (gpointer data)
{
  ((Editor *) data)->on_widget_creation ();
  return false;
}


void Editor::on_widget_creation ()
{
  // Scroll to the locaton of the cursor.
  GtkTextIter iter;
  gtk_text_buffer_get_iter_at_mark (textbuffer_text, &iter, gtk_text_buffer_get_insert (textbuffer_text));
  screen_scroll_to_itererator (GTK_TEXT_VIEW (textview_text), &iter);
}


void Editor::on_vpaned_editor_size_allocate (GtkWidget *widget, GdkRectangle *allocation, gpointer user_data)
{
  ((Editor *) user_data)->on_vpaned_editor_size_allocated (allocation);
}


void Editor::on_vpaned_editor_size_allocated (GdkRectangle *allocation)
{
  // Called when a size is allocated to the vpaned. If the size changed, it 
  // re-calculates the height of the Bible notes window.
  if (allocation->height != vpaned_previous_height) {
    vpaned_previous_height = allocation->height;
    bible_notes_window_set_size ();
  }
  // If the handle of the vpaned is moved by the user, it stores the new
  // size.
  bible_notes_window_store_size ();
}


void Editor::bible_notes_window_set_size ()
// This sets the height of the window displaying the Bible notes.
// If there are no notes, the height is zero.
// Else a percentage of the height of the whole Editor is used.
{
  GtkTextIter iter1, iter2;
  gtk_text_buffer_get_start_iter (textbuffer_notes, &iter1);
  gtk_text_buffer_get_end_iter (textbuffer_notes, &iter2);
  guint position = vpaned->allocation.height;
  if (!gtk_text_iter_equal (&iter1, &iter2)) {
    extern Settings * settings;
    ProjectConfiguration * projectconfig = settings->projectconfig (project);
    position = (guint) ((100 - projectconfig->vpane_biblenotes_percentage_get ()) * position / 100);
  }
  bible_notes_window_setting_size = true;
  gtk_paned_set_position (GTK_PANED (vpaned), position);
  bible_notes_window_setting_size = false;
}


void Editor::bible_notes_window_store_size ()
// If the handle of the paned is moved by the user,
// save the value in the configuration.
{
  // If the handle's position is set programmatically, bail out.
  if (bible_notes_window_setting_size) return;

  // If the bible notes buffer is empty, bail out too.
  GtkTextIter iter1, iter2;
  gtk_text_buffer_get_start_iter (textbuffer_notes, &iter1);
  gtk_text_buffer_get_end_iter (textbuffer_notes, &iter2);
  if (gtk_text_iter_equal (&iter1, &iter2)) return;
  
  // Get handle position. Bail out if it didn't change.
  unsigned int handle_position = gtk_paned_get_position (GTK_PANED (vpaned));
  if (handle_position == vpaned_previous_handle_position) return;
  vpaned_previous_handle_position = handle_position;

  // Store the new position in the configuration, in percentages.    
  unsigned int vpaned_height = vpaned->allocation.height;
  unsigned int percentage = vpaned_height - handle_position;
  percentage = percentage * 100 / vpaned_height;
  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  projectconfig->vpane_biblenotes_percentage_set (percentage);
}
