/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
*/


#include "libraries.h"
#include <glib.h>
#include "print_parallel_references.h"
#include "utilities.h"
#include "bible.h"
#include "usfm.h"
#include "usfmtools.h"
#include "pdfviewer.h"
#include "xmlutils.h"
#include "paper.h"
#include "formatter.h"
#include "xmlfo-utils.h"
#include "constants.h"
#include "gwrappers.h"
#include "gtkwrappers.h"
#include "directories.h"
#include "fonts.h"
#include "notecaller.h"
#include "xslfofootnote.h"
#include "xslfoendnote.h"
#include "xslfoxref.h"
#include "mapping.h"
#include "portion_utils.h"
#include "projectutils.h"
#include "progresswindow.h"
#include "books.h"
#include "unixwrappers.h"
#include "settings.h"
#include "textreplacement.h"


void view_parallel_references_pdf (const ustring& main_project,
                                   vector <ustring> * extra_projects,
                                   vector <Reference> references,
                                   bool keep_verses_together_within_page,
                                   vector<ustring> * remarks,
                                   bool highlight)
/*
Formats the references in "references", and highlights all words in
"session->highlights*" and shows them in a pdf viewer.
There is a pointer to a list of "remarks". Any remarks will be printed first.
*/
{
  // Log.
  gw_message ("Printing parallel references");

  // Configuration
  extern Settings * settings;
  
  // Progress system.
  ProgressWindow progresswindow ("Printing Parallel References", false);  
  progresswindow.set_text ("Collecting verses");
  progresswindow.set_iterate (0, 1, references.size());
  
  // Prepare for the inline text markers.
  ProjectConfiguration * projectconfig = settings->projectconfig (main_project);
  Usfm usfm (projectconfig->stylesheet_get());
  UsfmInlineMarkers usfm_inline_markers (usfm);

  // Prepare for leaving out footnotes, endnotes and crossreferences.
  XslFoFootnote xslfofootnote (usfm, false);
  XslFoEndnote xslfoendnote (usfm, false);
  XslFoXref xslfoxref (usfm, false);
  
  // Prepare for mapping.
  Mapping mapping (projectconfig->versification_get(), 0);
  
  // Storage for final xsl-fo file.
  vector<ustring> xslfo_lines;

  // Start building the xml-fo file.
  // The various elements of this file are created using objects.
  // When the object goes out of scope, it writes the closing elements.
  {
    XmlFoRoot xmlforoot (&xslfo_lines);
    {
      XmlFoLayoutMasterSet layoutmasterset (&xslfo_lines, false);
    }      
    {
      XmlFoPageSequence pagesequence (&xslfo_lines, false);
      {
        XmlFoStaticContent staticcontent (&xslfo_lines);
      }
      {
        XmlFoFlow flow (&xslfo_lines);
        {
          // Start off with inserting any remarks.
          if (remarks) {
            for (unsigned int r = 0; r < remarks->size(); r++) {
              xslfo_lines.push_back ("<fo:block space-before=\"2mm\" space-before.conditionality=\"retain\">");
              xslfo_lines.push_back (remarks->at (r));
              xslfo_lines.push_back ("</fo:block>");
            }
          }

          // Some variables to avoid excessive session database access during
          // highlighting.
          vector<bool> highlight_casesensitives;
          vector<ustring> highlight_words;
          if (highlight) {
            for (unsigned int hl = 0; hl < settings->session.highlights.size(); hl++) {
              highlight_casesensitives.push_back (settings->session.highlights[hl].casesensitive);
              highlight_words.push_back (settings->session.highlights[hl].word);
            }
          }
          
          // All the projects to be put in this parallel Bible, together with
          // their related information, like mapping, fonts.
          vector<ustring> project_s_raw;
          vector<ustring> project_names;
          vector<Mapping> mapping_s;
          vector<ustring> fonts;
          vector<unsigned int> fontsizes;
          vector<unsigned int> lineheights;
          if (extra_projects) project_s_raw = * extra_projects;
          for (unsigned int i = 0; i < project_s_raw.size(); i++) {
            ProjectConfiguration * projectconfig = settings->projectconfig (project_s_raw[i]);
            project_names.push_back(project_s_raw[i]);
            Mapping mapping (projectconfig->versification_get(), 0);
            mapping_s.push_back (mapping);
            PrintingFonts printingfonts (project_s_raw[i]);
            fonts.push_back (printingfonts.printing_families_comma_separated ());
            fontsizes.push_back (projectconfig->printing_font_size_get());
            lineheights.push_back (projectconfig->printing_line_height_get());
          }
          // Produce chunks for the xsl-fo file for all references.
          for (unsigned int rf = 0; rf < references.size(); rf++) {
            // Update progress bar.
            progresswindow.iterate();
            // Add the block to the xsl-fo file.
            // Next line has keep-together.within-page="always", rather than
            // keep-together="always", as the latter one causes things to be 
            // kept together, in in a line, which causes the line to overflow
            // the right margin.
            if (keep_verses_together_within_page)
              xslfo_lines.push_back ("<fo:block keep-together.within-page=\"always\">");
            else
              xslfo_lines.push_back ("<fo:block>");
            // XSLFormatter is better than FOP in that it does honour space conditionality,
            // which is initially set at "discard" for the beginning of a 
            // reference area, as here. So to get the distance between the 
            // lines right, this is inserted: space-before.conditionality="retain".
            xslfo_lines.push_back ("  <fo:block space-before=\"2mm\" space-before.conditionality=\"retain\">");
            xslfo_lines.push_back (references[rf].human_readable (""));
            xslfo_lines.push_back ("  </fo:block>");
            // Map this verse to the original, that is, to Hebrew or Greek.
            vector<int> hebrew_greek_chapters;
            vector<int> hebrew_greek_verses;
            mapping.book_change (references[rf].book);
            mapping.me_to_original (references[rf].chapter, references[rf].verse, hebrew_greek_chapters, hebrew_greek_verses);
            // Get verse text for each version.
            for (unsigned int vsn = 0; vsn <= project_names.size(); vsn++) {
              // Get the verse text.
              ustring line;
              if (vsn == 0) {
                // First version.
                line = project_retrieve_verse (main_project, references[rf].book, references[rf].chapter, references[rf].verse);
              } else {
                // Other versions. 
                // Get mapped chapters / verses.
                vector<int> mychapters;
                vector<int> myverses;
                mapping_s[vsn - 1].book_change (references[rf].book);
                mapping_s[vsn - 1].original_to_me (hebrew_greek_chapters, hebrew_greek_verses, mychapters, myverses);
                // Get text of any of the mapped verses.
                for (unsigned int mp = 0; mp < mychapters.size(); mp++) {
                  // Get the verse and add it to our container.
                  if (!line.empty()) 
                    line.append ("\n");
                  line.append (project_retrieve_verse (project_names [vsn - 1], references[rf].book, mychapters[mp], convert_to_string (myverses[mp])));
                }
              }
              // Do text replacements on this line
              text_replacement (line);
              // Take out verse number if it is the desired verse number, else leave it in.
              usfm_extract_marker (line);
              if (line.find (references[rf].verse) == 0)
                line.erase (0, references[rf].verse.length() + 1);
              // Take out footnotes, endnotes, crossreferences.
              xslfofootnote.transform (NULL, line);
              xslfoendnote.transform (NULL, line, NULL);
              xslfoxref.transform (NULL, line);              

              // Positions in the line, and lengths to highlight.
              vector <size_t> highlight_positions;
              vector <size_t> highlight_lengths;
              // Go through all the words to highlight.
              for (unsigned int i2 = 0; i2 < highlight_casesensitives.size(); i2++) {
                // Word to highlight
                ustring highlightword;
                if (highlight_casesensitives[i2]) 
                  highlightword = highlight_words[i2];
                else 
                  highlightword = highlight_words[i2].casefold ();
                // Variabele s holds a shadow string.
                ustring s;
                if (highlight_casesensitives[i2]) 
                  s = line;
                else 
                  s = line.casefold ();
                // Find positions for highlighting.
                size_t offposition = s.find (highlightword);
                while (offposition != string::npos) {
                  // Store position and length.
                  highlight_positions.push_back (offposition);
                  highlight_lengths.push_back (highlightword.length());
                  // Look for possible next word to highlight.
                  offposition = offposition + highlightword.length () + 1;
                  // There is something like a bug in s.find. If the offposition
                  // is greater than the length of s, then s.find will return
                  // a value below offposition, instead of string::npos as is
                  // expected. Workaround.
                  if (offposition > s.length())
                    break;
                  offposition = s.find (highlightword, offposition);
                }
              }
              // Sort the positions from small to bigger.
              xml_sort_positions (highlight_positions, highlight_lengths);
              // Combine overlapping positions.
              xml_combine_overlaps (highlight_positions, highlight_lengths);
              
              // Change <, > and & to their corresponding entities.
              xml_handle_entities (line, &highlight_positions);
              
              // Insert the code for highlighting.
              xml_fo_insert_emphasis (line, highlight_positions, highlight_lengths);
              // In case this prints changes, deal with strike-through and bold.
              replace_text (line, STRIKE_THROUGH_BEGIN, "<fo:inline text-decoration=\"line-through\">");
              replace_text (line, STRIKE_THROUGH_END, "</fo:inline>");
              replace_text (line, BOLD_BEGIN, "<fo:inline font-weight=\"bold\">");
              replace_text (line, BOLD_END, "</fo:inline>");
              
              // Deal with inline markers, so that they do not appear in the output
              // as markers, but format the text instead.
              usfm_handle_inline_text (line, &usfm_inline_markers, NULL, imXslFo, NULL);
              // This line has font information for each additional project. 
              ustring blockline;
              {
                blockline = "  <fo:block";
                if (vsn > 0) {
                  blockline.append (" font-family=\"" + fonts[vsn-1] + "\" font-size=\"" + convert_to_string (fontsizes[vsn-1]) + "pt\"");
                  if (lineheights[vsn-1] != 100)
                    blockline.append (" line-height=\"" + convert_to_string (lineheights[vsn-1]) + "%\"");
                }                  
                blockline.append (" text-indent=\"-5mm\">");
              }
              // We try to somehow show paragraphs also, basically.
              {
                ParseLine parseline (line);
                line.clear();
                for (unsigned int ln = 0; ln < parseline.lines.size(); ln++) {
                  ustring marker = usfm_extract_marker (parseline.lines[ln]);
                  if (!marker.empty()) {
                    if (usfm.is_starting_paragraph (marker)) {
                      if (ln > 0) {
                        line.append ("</fo:block>\n");
                        line.append (blockline);
                        line.append ("\n");
                      }
                    }
                  }
                  line.append (parseline.lines[ln]);
                  line.append ("\n");
                }
              }
              // Extract any marker left in the line.
              {
                ustring marker;
                size_t position, length;
                bool opening_marker;
                while (usfm_search_marker (line, marker, position, length, opening_marker)) {
                  line.erase (position, length);
                }
              }
              // Add the block opener.
              if (!project_names.empty()) xslfo_lines.push_back (blockline);
              // Add the project name, but only when there are more than one.
              if (!project_names.empty()) {
                ustring project;
                if (vsn == 0)
                  project = main_project;
                else
                  project = project_names[vsn - 1];
                xslfo_lines.push_back ("<fo:inline font-size=\"70%\" font-weight=\"bold\">" + project + "</fo:inline> ");
              }
              // Add the verse text itself.
              xslfo_lines.push_back (line);
              // Close the block.
              if (!project_names.empty()) xslfo_lines.push_back ("  </fo:block>");
            }
            // Close main block.
            xslfo_lines.push_back ("</fo:block>");
          }
        }
      }
    }
  }
  
  // Make a temporary directory where to put the working files and the resulting
  // .pdf file.information. This directory is not removed, because the pdf viewer 
  // needs the .pdf file to be there during viewing or printing.
  string working_directory = gw_build_filename(directories_get_temp(), "parallelreferences");
  gw_mkdir_with_parents (working_directory);
  // Produce filename of .fo file.
  string fofilename = gw_build_filename(working_directory, "document.fo");
  // Write the document.
  write_lines (fofilename, xslfo_lines);
  // Hide progeressbar.
  progresswindow.hide ();
  // Store the additional projects to print.
  vector<ustring> additional_projects;
  if (extra_projects) additional_projects = * extra_projects;
  settings->session.additional_printing_projects = additional_projects;
  // Convert the xsl-fo document to .pdf.
  NoteCaller dummy (nntNumerical, "", false);
  string pdffilename = gw_build_filename(working_directory, "document.pdf");
  bool conversion_result = formatter_convert_to_pdf (fofilename, pdffilename, &dummy, &dummy);
  // View the .pdf document.
  pdfviewer (pdffilename);
  // Give message if there were errors.
  if (!conversion_result) {
    string message = "The formatter had some trouble.\n"
                     "See menu Help - System log for more details.\n"
                     "See the helpfile for a possible solution.";
    gtkw_dialog_error (NULL, message);
  };
  // Log: ready.
  gw_message ("Ready printing the parallel references");
}
