/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "resource_viewer.h"
#include "directories.h"
#include "gwrappers.h"
#include "unixwrappers.h"
#include "utilities.h"
#include "books.h"
#include "settings.h"
#include "htmlbrowser.h"
#include "html.h"


ustring resource_viewer_directory ()
{
  return gw_build_filename (directories_get_temp (), "resource_viewer");  
}


void resource_viewer_ensure_directory ()
// Ensures that the working directory for the resource viewer is there.
{
  ustring directory = resource_viewer_directory();
  if (!g_file_test (directory.c_str(), G_FILE_TEST_IS_DIR)) {
    gw_mkdir_with_parents (directory);
  }
}


ustring resource_viewer_reference_html_filename ()
{
  ustring filename ("bibledit_resource_viewer_reference.html");
  return filename;
}


ustring resource_viewer_main_html_filename ()
{
  ustring filename ("bibledit_resource_viewer_main.html");
  return filename;
}


void resource_viewer_send_reference (Reference reference)
// Writes the Reference to the .html file so it can be picked up by the browser.
{
  // Ensure that the working directory is there.
  resource_viewer_ensure_directory ();
  
  // Load the .html file from the templates.
  ustring template_filename = gw_build_filename (directories_get_package_data (), resource_viewer_reference_html_filename ());
  ReadText rt (template_filename, true, false);
  
  // Write the reference to it.
  ustring reftext = resource_viewer_produce_anchor (reference.book, reference.chapter, convert_to_int (number_in_string (reference.verse)));
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    replace_text (rt.lines[i], "Matthew_1_1", reftext);
  }

  // Write the modified .html file to disk.
  ustring html_filename = gw_build_filename (resource_viewer_directory (), resource_viewer_reference_html_filename ());
  write_lines (html_filename, rt.lines);
}


void resource_viewer_open_page (const ustring& url)
// Opens a verse-synchronized resource in the browser.
{
  // Ensure that the .html file with the reference is there.
  extern Settings * settings;
  Reference reference (settings->genconfig.book_get (), convert_to_int (settings->genconfig.chapter_get ()), settings->genconfig.verse_get ());
  resource_viewer_send_reference (reference);
  
  // Load the .html file from the templates.
  ustring template_filename = gw_build_filename (directories_get_package_data (), resource_viewer_main_html_filename ());
  ReadText rt (template_filename, true, false);

  // Write the url to it. And the title.
  ustring title = html_extract_title (url);
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    replace_text (rt.lines[i], "king_james_bible_gospel_of_matthew.html", url);
    replace_text (rt.lines[i], "Bibledit Resources Viewer", title);
  }

  // Write the modified .html file to disk. Include the url in the name.
  ustring html_filename = gw_build_filename (resource_viewer_directory (), resource_viewer_main_html_filename ());
  ustring url_part (url);
  replace_text (url_part, "/", "_");
  html_filename.append (url_part);
  write_lines (html_filename, rt.lines);
  
  // Open the file in the html browser.
  htmlbrowser (html_filename, false);
}


ustring resource_viewer_produce_anchor (unsigned int book, unsigned int chapter, unsigned int verse)
// Produces the anchor, e.g.: Matthew_1_1 or Song_of_Solomon_2_2.
{
  ustring anchor (books_id_to_english (book) + "_" + convert_to_string (chapter) + "_" + convert_to_string (verse));
  replace_text (anchor, " ", "_");
  return anchor;
}


ResourceViewerStarter::ResourceViewerStarter (int dummy)
/*
This object starts the open resources in the Bibledit Resource Viewer.
It starts them one after the other, with a timeout in between.
The time between the first resource and the following ones is a bit longer
to allow the browser to start.
*/
{
  extern Settings * settings;
  urls = settings->genconfig.resources_viewer_urls_get ();
  pointer = 0;
  if (!urls.empty())
    g_timeout_add (1000, GSourceFunc (on_timeout), gpointer (this));
}


ResourceViewerStarter::~ResourceViewerStarter ()
{
}


bool ResourceViewerStarter::on_timeout (gpointer data)
{
  return ((ResourceViewerStarter *) data)->timeout();
}


bool ResourceViewerStarter::timeout ()
{
  // Open the first resource, allowing the browser to start.
  if (pointer == 0) {
    open_resource ();
  }
  
  // Increment pointer. If the pointer is past a certain limit, 
  // open the next resources, one by one.
  pointer++;
  if (pointer > 5) {
    open_resource ();
  }
  
  // If no resources are left, destroy the object.
  if (urls.empty()) {
    delete this;
    return false;    
  }
  
  // Continue with next resource.
  return true;
}


void ResourceViewerStarter::open_resource ()
// Do the administration of opening a resource.
{
  if (urls.empty()) return;
  resource_viewer_open_page (urls[0]);
  urls.erase (urls.begin());
}
