/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "subversion.h"
#include "settings.h"
#include "gwrappers.h"
#include "date_time_utils.h"
#include "utilities.h"
#include "projectutils.h"
#include "books.h"
#include "unixwrappers.h"
#include "directories.h"
#include "shell.h"
#include "gtkwrappers.h"
#include "progresswindow.h"
#include "books.h"
#include "vacuum.h"
#include "ipc.h"
#include <libxml/xmlwriter.h>
#include <libxml/xmlreader.h>


vector <SubversionTask> subversiontasks;


SubversionTask::SubversionTask (SubversionTaskType task_in, const ustring& project_in, unsigned int book_in, unsigned int chapter_in, unsigned int failures_in)
{
  task = task_in;
  project = project_in;
  book = book_in;
  chapter = chapter_in;
  failures = failures_in;
}


ustring subversion_repository_local (const ustring& project, bool includeprefix)
{
  ustring repository;
  if (includeprefix) repository = "file://";
  repository.append (gw_build_filename (directories_get_subversion (), project));
  return repository;
}


void subversion_schedule (SubversionTaskType task, const ustring& project, 
                          unsigned int book, unsigned int chapter)
// This schedules a subversion task.
{
  SubversionTask subversiontask (task, project, book, chapter, 0);
  subversiontasks.push_back (subversiontask);
}


void subversion_initial_check_all (bool gui)
// Does initial checks on the subversion repositories.
{
  // Do all projects.
  vector<ustring> projects = projects_get_all ();
  for (unsigned int i = 0; i < projects.size(); i++) {
    subversion_initial_check_project (projects[i]);
  }
}


void subversion_initial_check_project (const ustring& project)
// Does initial checks on the subversion repositories.
// Upgrades if needed.
{
  // Skip if repository exists.
  ustring directory = subversion_repository_local (project, false);
  if (g_file_test (directory.c_str(), G_FILE_TEST_IS_DIR))
    return;

  // Progress.
  cout << "Creating subversion repository for " << project << endl;
  
  // Create the repository.
  // Older versions of svnadmin create do not default to fsfs, so pass this argument always.  
  {
    GwSpawn spawn ("svnadmin");
    spawn.arg ("create");
    spawn.arg ("--fs-type");
    spawn.arg ("fsfs");
    spawn.arg (directory);
    spawn.run ();
  }
  
  // Schedule project.
  // This will switch the remote repository off, so it will use the local one.
  subversion_schedule (sttInitializeProject, project, 0, 0);
}


void subversion_store_chapter (const ustring& project, unsigned int book, unsigned int chapter)
{
  subversion_schedule (sttStoreChapter, project, book, chapter);
}


void subversion_move_project (const ustring& project, const ustring& newproject)
{
  // Update all pending operations for this project.
  for (unsigned int i = 0; i < subversiontasks.size (); i++) {
    if (subversiontasks[i].project == project)
      subversiontasks[i].project = newproject;
  }
  // Move the repository.
  ustring repositorydir = subversion_repository_local (project, false);
  ustring newrepositorydir = subversion_repository_local (newproject, false);
  unix_mv (repositorydir, newrepositorydir);
  // A svn switch --relocate is needed if the repository is local.
  // A remote repository is left as it is.
  ustring url = subversion_project_get_url (project);
  if (url.find ("file://") == 0) {
    ustring newurl = gw_path_get_dirname (url);
    newurl = gw_build_filename (newurl, newproject);
    GwSpawn spawn ("svn");
    spawn.workingdirectory (project_data_directory_project (project));
    spawn.arg ("switch");
    spawn.arg ("--relocate");
    spawn.arg (url);
    spawn.arg (newurl);
    spawn.arg ("--non-interactive");
    spawn.run ();
  }
}


void subversion_remove_book (const ustring& project, unsigned int book)
{
  // Remove all pending operations for this project/book.
  vector <SubversionTask> newtasks;
  for (unsigned int i = 0; i < subversiontasks.size(); i++) {
    if ((subversiontasks[i].project != project) || (subversiontasks[i].book != book))
      newtasks.push_back (subversiontasks[i]);  
  }
  subversiontasks = newtasks;
  // Using svn remove the book.
  GwSpawn spawn ("svn");
  spawn.workingdirectory (project_data_directory_project (project));
  spawn.arg ("rm");
  spawn.arg (books_id_to_english (book));
  spawn.arg ("--force");
  spawn.arg ("--non-interactive");
  spawn.run ();
  // Schedule project commit.
  subversion_schedule (sttCommitProject, project, 0, 0);
}


void subversion_remove_chapter (const ustring& project, unsigned int book, unsigned int chapter)
{
  // Remove all pending operations for this project/book/chapter.
  vector <SubversionTask> newtasks;
  for (unsigned int i = 0; i < subversiontasks.size(); i++) {
    if ((subversiontasks[i].project != project) || (subversiontasks[i].book != book) || (subversiontasks[i].chapter != chapter))
      newtasks.push_back (subversiontasks[i]);  
  }
  subversiontasks = newtasks;
  // Using svn remove the chapter.
  GwSpawn spawn ("svn");
  spawn.workingdirectory (project_data_directory_book (project, book));
  spawn.arg ("rm");
  spawn.arg (convert_to_string (chapter));
  spawn.arg ("--force");
  spawn.arg ("--non-interactive");
  spawn.run ();
  // Schedule a project commit.
  subversion_schedule (sttCommitProject, project, 0, 0);
}


void subversion_remove_project (const ustring& project)
{
  // Remove all pending operations for this project.
  subversion_erase_tasks_project (project);
  // Remove the local repository, and leave the possible remote one.
  ustring repositorydir = subversion_repository_local (project, false);
  unix_rmdir (repositorydir);
}


void subversion_cleanup ()
/*
When Subversion modifies your working copy (or any information within .svn), 
it tries to do so as safely as possible. Before changing the working copy, 
Subversion writes its intentions to a log file. Next it executes the commands in 
the log file to apply the requested change, holding a lock on the relevant part 
of the working copy while it works — to prevent other Subversion clients from 
accessing the working copy in mid-change. Finally, Subversion removes the log 
file. Architecturally, this is similar to a journaled filesystem. If a Subversion 
operation is interrupted (if the process is killed, or if the machine crashes, 
for example), the log files remain on disk. By re-executing the log files, 
Subversion can complete the previously started operation, and your working copy 
can get itself back into a consistent state.
And this is exactly what svn cleanup does: it searches your working copy and 
runs any leftover logs, removing working copy locks in the process. If 
Subversion ever tells you that some part of your working copy is “locked”, then 
this is the command that you should run. Also, svn status will display an L 
next to locked items.

svn cleanup can take a good while to complete. Therefore we only do the 
project that is open in the editor.

*/
{
  // Find the open project. Bail out if there is none.
  extern Settings * settings;
  ustring project = settings->genconfig.project_get();
  if (project.empty()) return;
  
  /*
  We sometimes discover that the svn operations don't work properly, 
  because some directories appear to be missing.
  See an example of such a failure:
    Storing Mark 1
    svn: Commit failed (details follow):
    svn: Your .svn/tmp directory may be missing or corrupt; run 'svn cleanup' and try again
    svn: Can't open file '/home/joe/.bibledit/projects/ERV/data/Mark/1/.svn/tmp/entries': No such file or directory
  Not only commit fails in such cases, it has been found, but also svn cleanup.
  Therefore manually check on these directories before running the cleanup 
  command.  
  */
  vector<unsigned int> books = project_get_books (project);
  for (unsigned int bk = 0; bk < books.size(); bk++) {
    vector<unsigned int> chapters = project_get_chapters (project, books[bk]);
    for (unsigned int ch = 0; ch < chapters.size(); ch++) {
      ustring directory = project_data_directory_chapter (project, books[bk], chapters[ch]);
      ustring tmp = gw_build_filename (directory, ".svn", "tmp");
      if (!g_file_test (tmp.c_str(), G_FILE_TEST_IS_DIR)) {
        gw_mkdir_with_parents (tmp);
      }
    }
  }

  // Run the svn cleanup.
  GwSpawn spawn ("svn");
  spawn.workingdirectory (project_data_directory_project (project));
  gw_message ("Cleaning up " + project_data_directory_project (project));
  spawn.arg ("cleanup");
  spawn.async ();
  spawn.run ();
}


void subversion_get_tasks (vector <unsigned int>& tasks, vector <ustring>& projects,
                           vector <unsigned int>& books, vector <unsigned int>& chapters,
                           vector <unsigned int>& fails)
{
  for (unsigned int i = 0; i < subversiontasks.size(); i++) {
    tasks.push_back    (subversiontasks[i].task);
    projects.push_back (subversiontasks[i].project);
    books.push_back    (subversiontasks[i].book);
    chapters.push_back (subversiontasks[i].chapter);
    fails.push_back    (subversiontasks[i].failures);
  }
}


void subversion_task_human_readable (unsigned int task, const ustring& project,
                                     unsigned int book, unsigned int chapter,
                                     unsigned int fail,
                                     gchar * & human_readable_task,
                                     ustring& human_readable_description,
                                     ustring& human_readable_status)
{
  // Describe the task.
  switch ((SubversionTaskType) task) {
    case sttInitializeProject:
    {
      human_readable_task = "Initialize project";
      break;
    }
    case sttCommitProject:
    {
      human_readable_task = "Commit project";
      break;
    }
    case sttStoreChapter:
    {
      human_readable_task = "Store chapter";
      break;
    }
    case sttUpdateProject:
    {
      human_readable_task = "Update project";
      break;
    }
    case sttReopenProject:
    {
      human_readable_task = "Reopen project";
      break;
    }
    case sttOpenChapter:
    {
      human_readable_task = "Open chapter";
      break;
    }
  }
  // Give description of task.
  switch ((SubversionTaskType) task) {
    case sttStoreChapter: 
    case sttOpenChapter: 
    {
      human_readable_description = books_id_to_english (book) + " " + convert_to_string (chapter);
      break;
    }
    case sttUpdateProject:
    case sttInitializeProject:
    case sttCommitProject:
    case sttReopenProject:
    {
      break;
    }
  }
  // Describe status.  
  if (fail == 0) {
    human_readable_status = "Pending";
  } else {
    human_readable_status = "Failed " + convert_to_string (fail) + " times, will retry";
  }
}


void subversion_erase_tasks_project (const ustring& project)
// Remove all pending operations for the project.
{
  vector <SubversionTask> newtasks;
  for (unsigned int i = 0; i < subversiontasks.size(); i++) {
    if (subversiontasks[i].project != project)
      newtasks.push_back (subversiontasks[i]);  
  }
  subversiontasks = newtasks;
}


int subversion_count_tasks_project (const ustring& project)
{
  int count = 0;
  for (unsigned int i = 0; i < subversiontasks.size(); i++) {
    if (subversiontasks[i].project == project)
      count++;
  }
  return count;
}


void subversion_open_chapter_for_editing (const ustring& project, unsigned int book, unsigned int chapter)
/*
Do the remote repository administration of opening a chapter for editing.

Generally speaking this means running "svn update" on the chapter,
and taking action depending on the output of that command.

All this is not done when just retrieving the chapter's data.
This would slow Bibledit down too much. 
It should be done only when a chapter is opened for editing.

*/
{
  // Bail out if our repository is local.
  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  if (!projectconfig->repository_remote_get ())
    return;
  
  // Update the chapter from the remote repository, and take action depending
  // on the outcome.
  bool dummy;
  SubversionUpdateType outcome = subversion_update_chapter_remote (project, book, chapter, dummy);
  switch (outcome) {
    case sutOk:
    case sutAdded:
    {
      // No update was done or the chapter was updated.
      // Actions: None.
      break;
    }
    case sutUpdated:
    case sutReplaced:
    case sutMerged:
    case sutDeleted: 
    {
      // The chapter was updated, added, replaced, or merged. 
      // Or the chapter was deleted. Actions:
      // - schedule a re-open project.
      subversion_schedule (sttReopenProject, project, 0, 0);
      break;
    }
    case sutConflict:
    case sutError:    
    {
      // There are conflicting changes that could not be automatically merged.
      // Or there was an error updating the project. 
      // Usually this means that the server is down or the user is offline.
      // Actions: 
      // - schedule an update for the project so it will keep trying.
      subversion_schedule (sttUpdateProject, project, 0, 0);
      break;
    }
  }
}


bool subversion_execute_update_project (const ustring& project)
/*
Runs svn update on the whole project.
It will take the right action on all chapters that will be marked.
See subversion_update_chapter_remote () for possible output of svn update.
*/
{
  // Log
  gw_message ("Updating project " + project);
  
  // Initialize variables.
  bool ran_okay = false;
  Reference reference = subversion_execute_retrieve_reference ();
  ustring directory = project_data_directory_project (project);

  // Run the svn update.
  GwSpawn spawn ("svn");
  spawn.workingdirectory (directory);
  spawn.arg ("update");
  spawn.arg ("--non-interactive");
  spawn.read ();
  spawn.run ();
  if (spawn.exitstatus == 0) {
    ran_okay = true;
    for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
      SubversionUpdateType result = sutOk;
      // Results look like, e.g: "U    3 John/1/data".
      ustring fragment = spawn.standardout[i].substr (0, 2);
      if (fragment == "U ") result = sutUpdated;
      if (fragment == "A ") result = sutAdded;
      if (fragment == "D ") result = sutDeleted;
      if (fragment == "R ") result = sutReplaced;
      if (fragment == "G ") result = sutMerged;
      if (fragment == "C ") result = sutConflict;
      // Only proceed if we've enough text.
      ustring bookchapterfragment = spawn.standardout[i];
      size_t dataposition = bookchapterfragment.find ("/data");
      if (dataposition != string::npos) 
        bookchapterfragment.erase (dataposition, 10);
      if (bookchapterfragment.length() > 3) {
        bookchapterfragment.erase (0, 2);
        bookchapterfragment = trim (bookchapterfragment);
        ustring fragment1 = gw_path_get_dirname (bookchapterfragment);
        ustring fragment2 = gw_path_get_basename (bookchapterfragment);
        unsigned int book = books_english_to_id (fragment1);
        unsigned int chapter = convert_to_int (fragment2);
        if (book) {
          // We've got a book and chapter.
          subversion_update_chapter_remote_scheduler (project, book, chapter, result, &reference);
        } else {
          // We might have had something like "D   John". Handle that here.
          book = books_english_to_id (fragment2);
          if (book) {
            if (result == sutDeleted) {
              subversion_schedule (sttReopenProject, project, 0, 0);
            }
          }
        }
      }
    }      
  }    
  
  // Return whether success.
  return ran_okay;
}


SubversionUpdateType subversion_update_chapter_remote (const ustring& project, unsigned int book, unsigned int chapter, bool& ran_okay)
/*
Runs svn update on one chapter in the permanent checkout directory.
One chapter will be faster than the whole project, hence it is done here.
It returns the SubversionUpdateType, i.e. what needs to be done.
Possible output of svn update:
U foo    File foo was updated (received changes from the server).
A foo    File or directory foo was added to your working copy.
D foo    File or directory foo was deleted from your working copy.
R foo    File or directory foo was replaced in your working copy; that is, 
         foo was deleted, and a new item with the same name was added. While 
         they may have the same name, the repository considers them to be 
         distinct objects with distinct histories.
G foo    File foo received new changes from the repository, but your local copy 
         of the file had your modifications. Either the changes did not intersect, 
         or the changes were exactly the same as your local modifications, so 
         Subversion has successfully merged the repository's changes into the 
         file without a problem.
C foo    File foo received conflicting changes from the server. The changes 
         from the server directly overlap your own changes to the file. No need 
         to panic, though. This overlap needs to be resolved by a human.
*/
{
  SubversionUpdateType result = sutError;
  ran_okay = false;
  ustring workingdirectory = project_data_directory_project (project);
  ustring chapterfragment = gw_build_filename (books_id_to_english (book), convert_to_string (chapter));
  if (g_file_test (gw_build_filename (workingdirectory, chapterfragment).c_str(), G_FILE_TEST_IS_DIR)) {
    GwSpawn spawn ("svn");
    spawn.workingdirectory (workingdirectory);
    spawn.arg ("update");
    spawn.arg (chapterfragment);
    spawn.arg ("--non-interactive");
    spawn.read ();
    spawn.run ();
    if (spawn.exitstatus == 0) {
      ran_okay = true;
      result = sutOk;
      if (!spawn.standardout.empty()) {
        // Results look like, e.g: "U    3 John/1/data".
        ustring fragment = spawn.standardout[0].substr (0, 2);
        if (fragment == "U ") result = sutUpdated;
        if (fragment == "A ") result = sutAdded;
        if (fragment == "D ") result = sutDeleted;
        if (fragment == "R ") result = sutReplaced;
        if (fragment == "G ") result = sutMerged;
        if (fragment == "C ") result = sutConflict;
      }      
    }    
  }
  return result;
}


void subversion_update_chapter_remote_scheduler (const ustring& project, unsigned int book, unsigned int chapter, SubversionUpdateType updatetype, Reference * editor_reference)
// Depending on the updatetype passed, this might schedule a task for this 
// chapter or execute it straightaway.
{
  extern Settings * settings;
  switch (updatetype) {
    case sutOk:       // Everything is fine.
    case sutError:    // There was an error. Will try again later on.
    {
      break;
    }
    case sutUpdated:  // Chapter has been updated, replaced, or merged.
    case sutReplaced: // Action: Normally no action, but if the editor shows
    case sutMerged:   // this project, book and chapter, chapter, reopen project.
    {
      if (settings->genconfig.project_get() == project) {
        if (editor_reference->book == book) {
          if (editor_reference->chapter == chapter) {
            subversion_schedule (sttReopenProject, project, 0, 0);
          }
        }
      }
      break;
    }
    case sutAdded:    // Chapter has been added. Signal to reopen project.
    case sutDeleted:  // Chapter has been deleted. Signal to reopen project.
    {
      subversion_schedule (sttReopenProject, project, 0, 0);
      break;
    }
    case sutConflict: // Chapter is in a conflicting state. Solve conflict.
    {
      subversion_resolve_conflict_chapter (project, book, chapter);
      break;
    }
  }
}


ustring subversion_project_get_url (const ustring& project)
// Gets the current URL of the repository, through svn info.
{
  ustring workingdirectory = project_data_directory_project (project);
  GwSpawn spawn ("svn");
  spawn.workingdirectory (workingdirectory);
  spawn.arg ("info");
  spawn.arg ("--non-interactive");
  spawn.read ();  
  spawn.run ();
  ustring url;
  for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
    if (spawn.standardout[i].find ("URL: ") == 0) {
      spawn.standardout[i].erase (0, 5);
      url = spawn.standardout[i];
    }
  }
  return url; 
}


void subversion_revert_to_internal_repository (const ustring& project)
// This reverts the repository to the internal one, if that is not yet the case.
{
  // If we're already using the internal repository, bail out.
  ustring url = subversion_project_get_url (project);
  if (url == subversion_repository_local (project, true))
    return;
  // Set in the configuration that we're using a local repository.
  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  projectconfig->repository_remote_set (false);
  // Remove the local repository directory, then run a project update on it.
  // This will do the right things.
  unix_rmdir (subversion_repository_local (project, false));
  subversion_initial_check_project (project);  
}


void subversion_resolve_conflict_chapter (const ustring& project, unsigned int book, unsigned int chapter)
// This solves a conflicting chapter.
{
  // Log message
  gw_message ("Resolving conflict for " + books_id_to_english (book) + " " + convert_to_string (chapter));

  // Directory of the chapter in the data.
  ustring directory = project_data_directory_chapter (project, book, chapter);

  extern Settings * settings;
  ProjectConfiguration * projectconfig = settings->projectconfig (project);
  SubversionConflictHandlingType conflicthandling = (SubversionConflictHandlingType) projectconfig->repository_conflict_get ();
  
  // Datafile name.
  ustring datafile = gw_build_filename (directory, "data");
  
  /* Read the datafile. This may look like the below example if there is a conflict:
  \c 1
  <<<<<<< .mine
  \v 1 my text
  =======
  \v 1 server's text
  >>>>>>> .r1251
  \v 2
  The first line between the markers is "mine", and the second line between
  the markers is the server's version, revision "r1251".
  */
  ReadText rt (datafile, true);
  
  // Set about to resolve the conflict.
  vector <ustring> newdata;
  bool withinmine = false;
  bool withinserver = false;
  for (unsigned int i = 0; i < rt.lines.size(); i++) {
    // Find out when we've a marker, no actual data.    
    bool minemarker = rt.lines[i] == subversion_mine_conflict_marker ();
    bool separatormarker = rt.lines[i] == "=======";
    bool servermarker = rt.lines[i].find (">>>>>>> .r") == 0;
    // Do conflict management only if we've actual data.
    bool takeit = false;
    if (!minemarker && !separatormarker && !servermarker) {
      if (withinmine) {
        if (conflicthandling == schtTakeMe) takeit = true;
      } else if (withinserver) {
        if (conflicthandling == schtTakeServer) takeit = true;
      } else {
        takeit = true;
      }   
    }
    if (takeit) {
      newdata.push_back (rt.lines[i]);
    } else {
    }
    // Set whether we're within a conflict, my data, or the server's data.
    // This setting applies to the next line of data.
    if (minemarker) {
      withinmine = true;
      withinserver = false;
    }
    if (separatormarker) {
      withinmine = false;
      withinserver = true;
    }
    if (servermarker) {
      withinmine = false;
      withinserver = false;
    }
  }    
  write_lines (datafile, newdata);
  
  // Set the conflict resolved.
  GwSpawn spawn ("svn");
  spawn.workingdirectory (directory);
  spawn.arg ("resolved");
  spawn.arg ("data");
  spawn.run ();

  // Schedule a commit.
  subversion_schedule (sttCommitProject, project, 0, 0);
}


bool subversion_log_extract_date_time (const ustring& line, int& year, int& month, int& day, int& hour, int& minute, int& second)
/* 
Reads a line of "svn log" and extract the date and time from it.
Typical lines:

------------------------------------------------------------------------
r1 | joe | 2006-12-25 07:47:15 +0200 (Mon, 25 Dec 2006) | 1 line

store book
------------------------------------------------------------------------

*/
{
  bool success = false;
  if (line.length() > 20) {
    if (line.substr (0, 1) == "r") {
      if (line.find ("|") != string::npos) {
        Parse parse (line, false, "|");
        if (parse.words.size() == 4) {
          ustring datetime = parse.words[2];
          if (datetime.length () > 20) {
            replace_text (datetime, "-", " ");
            replace_text (datetime, ":", " ");
            Parse parse2 (datetime);
            if (parse2.words.size() > 6) {
              year = convert_to_int (parse2.words[0]);
              month = convert_to_int (parse2.words[1]);
              day = convert_to_int (parse2.words[2]);
              hour = convert_to_int (parse2.words[3]);
              minute = convert_to_int (parse2.words[4]);
              second = convert_to_int (parse2.words[5]);
              success = true;
            }
          }          
        }
      }
    }
  }  
  return success;
}


unsigned int subversion_log_date_time_at_revision (const ustring& project, unsigned int revision)
// Get the date/time at the revision for the project.
{
  ustring workingdirectory = project_data_directory_project (project);
  GwSpawn spawn ("svn");
  spawn.workingdirectory (workingdirectory);
  spawn.arg ("log");
  spawn.arg ("-r");
  spawn.arg (convert_to_string (revision));
  spawn.arg ("--non-interactive");
  spawn.progress ("Accessing log", false);
  spawn.read ();
  spawn.run ();
  unsigned int seconds = 0;
  if (spawn.standardout.size() > 1) {
    for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
      int year, month, day, hour, minute, second;
      if (subversion_log_extract_date_time (spawn.standardout[i], year, month, day, hour, minute, second)) {
        int julian_day = date_time_julian_day_get_parse (year, month, day);
        seconds = date_time_julian_to_seconds (julian_day);
        seconds += 3600 * hour;
        seconds += 60 * minute;
        seconds += second;        
      }
    }  
  }
  return seconds;
}


ustring subversion_revision_date_time (unsigned int seconds)
// This turns the "seconds" into a date and time as used by subversion, e.g.
// {2006-01-05 10:15:01}
{
  unsigned int year, month, day;
  int hour, minute, second;
  int julian_day = date_time_seconds_to_julian (seconds);
  date_time_normal_get_year_month_day (julian_day, year, month, day);
  date_time_normal_get_hour_minute_second (seconds, hour, minute, second);
  ustring uyear, umonth, uday, uhour, uminute, usecond;
  uyear = convert_to_string (year);
  umonth = convert_to_string (month);
  if (umonth.length() == 1) umonth.insert (0, "0");
  uday = convert_to_string (day);
  if (uday.length() == 1) uday.insert (0, "0");
  uhour = convert_to_string (hour);
  if (uhour.length() == 1) uhour.insert (0, "0");
  uminute = convert_to_string (minute);
  if (uminute.length() == 1) uminute.insert (0, "0");
  usecond = convert_to_string (second);
  if (usecond.length() == 1) usecond.insert (0, "0");
  ustring datetime = "{";
  datetime.append (uyear + "-" + umonth + "-" + uday);
  datetime.append (" ");
  datetime.append (uhour + ":" + uminute + ":" + usecond);
  datetime.append ("}");
  return datetime;
}


void subversion_get_chapters_changed_since (const ustring& project, int seconds, vector <unsigned int>& books, vector <unsigned int>& chapters)
// This gives the books and chapters changed since "seconds".
{
  // Checkout the history project.
  // This contains the data of our project as it was at the date/time given.
  ustring history_project = project + "_as_it_was_at_" + convert_to_string (seconds);
  ustring history_project_data_directory = project_data_directory_project (history_project);
  gw_mkdir_with_parents (history_project_data_directory);
  {
    ustring workingdirectory = gw_path_get_dirname (history_project_data_directory);
    ustring repositorypath = subversion_project_get_url (project);
    ustring checkoutdirectory = gw_build_filename (workingdirectory, gw_path_get_basename (repositorypath));
    GwSpawn spawn ("svn");
    spawn.workingdirectory (workingdirectory);
    spawn.arg ("checkout");
    spawn.arg (repositorypath);
    spawn.arg ("--revision");
    spawn.arg (subversion_revision_date_time (seconds));
    spawn.arg ("--non-interactive");
    spawn.progress ("Retrieving data from history", false);
    spawn.run ();
    if (spawn.exitstatus == 0) {
      unix_rmdir (history_project_data_directory);
      unix_mv (checkoutdirectory, history_project_data_directory);
    } else {
      gtkw_dialog_error (NULL, "Failed to retrieve history");
      project_delete (history_project);  
      return;
    }
  }

  // Go through all books of the project.
  vector <unsigned int> projectbooks = project_get_books (project);
  for (unsigned int bk = 0; bk < projectbooks.size(); bk++) {
    // Go through all chapters.
    vector <unsigned int> projectchapters = project_get_chapters (project, projectbooks[bk]);
    for (unsigned int ch = 0; ch < projectchapters.size(); ch++) {
      // Get chapter data and compare it.
      // Add differing books/chapters to the list.
      ustring line_now;
      {
        vector <ustring> chapterdata = project_retrieve_chapter (project, projectbooks[bk], projectchapters[ch]);
        for (unsigned int i = 0; i < chapterdata.size(); i++)
          line_now.append (chapterdata[i]);
      }
      ustring line_history;
      {
        vector <ustring> chapterdata = project_retrieve_chapter (history_project, projectbooks[bk], projectchapters[ch]);
        for (unsigned int i = 0; i < chapterdata.size(); i++)
          line_history.append (chapterdata[i]);
      }
      if (line_now != line_history) {
        books.push_back (projectbooks[bk]);
        chapters.push_back (projectchapters[ch]);
      }      
    }
  }  
  
  // Clean up.  
  project_delete (history_project);  
}


void subversion_log_get_seconds_and_revisions (const vector<ustring>& log, vector <unsigned int>& seconds, vector <unsigned int>& revisions)
// Taking the lines in the "log", it retrieves a list of:
// seconds: containing the seconds since the Epoch, and
// revision: containing the revision numbers as used by Subversion.
{
  // Clear containers.
  seconds.clear();
  revisions.clear();

  // Go through the log.
  for (unsigned int i = 0; i < log.size(); i++) {
    // If we extract the date, it means we have a valid line here.
    int year, month, day, hour, minute, second;
    if (subversion_log_extract_date_time (log[i], year, month, day, hour, minute, second)) {
      // Calculate seconds.
      int julian_day = date_time_julian_day_get_parse (year, month, day);
      int sec = date_time_julian_to_seconds (julian_day);
      sec += 3600 * hour;
      sec += 60 * minute;
      sec += second;
      // Revision.
      ustring rev = log[i];
      rev.erase (0, 1);
      rev = number_in_string (rev);
      int revision = convert_to_int (rev);
      // Store.
      seconds.push_back (sec);
      revisions.push_back (revision);
    }    
  }
}


ustring subversion_job_filename ()
// Returns the filename of the xml files with the subversion jobs
{
  return gw_build_filename (directories_get_temp (), "subversionjobs.xml");
}


void subversion_command_initialize_subsystem ()
{
  // Load any tasks that were left undone on the previous shutdown.
  if (g_file_test (subversion_job_filename ().c_str(), G_FILE_TEST_IS_REGULAR)) {

    // Load file in memory and create the input buffer.
    gchar *contents;
    g_file_get_contents (subversion_job_filename ().c_str(), &contents, NULL, NULL);
    xmlParserInputBufferPtr inputbuffer;
    inputbuffer = xmlParserInputBufferCreateMem(contents, strlen (contents), XML_CHAR_ENCODING_NONE);

    // Create a reader.    
    xmlTextReaderPtr reader = xmlNewTextReader (inputbuffer, NULL);
    
    // Iterate through the nodes in streaming mode, loading tasks.
    if (reader) {
      int read;
      SubversionTask task (sttUpdateProject, "", 0, 0, 0);
      ustring text;
      while ((read = xmlTextReaderRead(reader) == 1)) {
        switch (xmlTextReaderNodeType (reader)) {
          case XML_READER_TYPE_ELEMENT:
          {
            break;
          }
          case XML_READER_TYPE_TEXT:
          {
            xmlChar * value = xmlTextReaderValue (reader);
            if (value) {
              text = (const char *) value;
              xmlFree(value);
            }
            break;
          }
          case XML_READER_TYPE_END_ELEMENT:
          {
            // Retrieve the name of the element and process it.
            xmlChar * element_name = xmlTextReaderName (reader);
            if (!xmlStrcmp(element_name, BAD_CAST "job")) {
              subversiontasks.push_back (task);
            }
            else if (!xmlStrcmp(element_name, BAD_CAST "task")) {
              task.task = (SubversionTaskType) convert_to_int (text);
            }
            else if (!xmlStrcmp(element_name, BAD_CAST "project")) {
              task.project = text;
            }
            else if (!xmlStrcmp(element_name, BAD_CAST "book")) {
              task.book = convert_to_int (text);
            }
            else if (!xmlStrcmp(element_name, BAD_CAST "chapter")) {
              task.chapter = convert_to_int (text);
            }
            else if (!xmlStrcmp(element_name, BAD_CAST "failures")) {
              task.failures = convert_to_int (text);
            }
            break;
          }
        }
      }
    }
  
    // Free memory.
    if (reader) xmlFreeTextReader(reader);
    if (inputbuffer) xmlFreeParserInputBuffer (inputbuffer);
    if (contents) free (contents);  
    
    // Remove the file.
    unlink (subversion_job_filename ().c_str());
  }
  
}


void subversion_command_finalize_subsystem ()
{
  // Send the shutdown command to bibledit-subversion.
  vector <ustring> payload;
  payload.push_back (" ");
  extern InterprocessCommunication * ipc;
  ipc->send (ipcstBibleditSubversion, ipcctSubversionShutdown, payload);

  // If there are still any subversion jobs left, save them to file
  // for loading and execution during the next run.
  if (!subversiontasks.empty()) {
    
    // Create a new XML buffer, to which the document will be written
    xmlBufferPtr buffer = xmlBufferCreate();
  
    // Create a new XmlWriter for memory, with no compression.
    xmlTextWriterPtr writer = xmlNewTextWriterMemory(buffer, 0);

    // Create first bit of the document.
    xmlTextWriterStartDocument(writer, NULL, "UTF-8", NULL);
  
    // Indent the output.
    xmlTextWriterSetIndent(writer, 1);
    
    // Write the main element.
    xmlTextWriterStartElement (writer, BAD_CAST "subversion-tasks");
    
    // Go through all available tasks and write them out.
    for (unsigned int i = 0; i < subversiontasks.size(); i++) {
      xmlTextWriterStartElement (writer, BAD_CAST "job");
      xmlTextWriterStartElement (writer, BAD_CAST "task");
      xmlTextWriterWriteFormatString (writer, "%d", subversiontasks[i].task);
      xmlTextWriterEndElement (writer);
      xmlTextWriterStartElement (writer, BAD_CAST "project");
      xmlTextWriterWriteFormatString (writer, "%s", subversiontasks[i].project.c_str());
      xmlTextWriterEndElement (writer);
      xmlTextWriterStartElement (writer, BAD_CAST "book");
      xmlTextWriterWriteFormatString (writer, "%d", subversiontasks[i].book);
      xmlTextWriterEndElement (writer);
      xmlTextWriterStartElement (writer, BAD_CAST "chapter");
      xmlTextWriterWriteFormatString (writer, "%d", subversiontasks[i].chapter);
      xmlTextWriterEndElement (writer);
      xmlTextWriterStartElement (writer, BAD_CAST "failures");
      xmlTextWriterWriteFormatString (writer, "%d", subversiontasks[i].failures);
      xmlTextWriterEndElement (writer);
      xmlTextWriterEndElement (writer);
    }
  
    // We could close any open elements by hand, but closing the document does it all.
    xmlTextWriterEndDocument (writer);
  
    // Flush the writer, else nothing gets transferred to the buffer.
    xmlTextWriterFlush (writer);

    // Write the lines to disk.
    ParseLine parseline ((const char *) buffer->content);
    write_lines (subversion_job_filename (), parseline.lines);

    // Free memory.
    if (writer) xmlFreeTextWriter (writer);
    if (buffer) xmlBufferFree (buffer);
  }  
}


bool subversion_execute_store_chapter (const ustring& project, unsigned int book, unsigned int chapter)
// Stores a chapter. Returns whether things worked out.
{
  // Log message
  gw_message ("Storing " + books_id_to_english (book) + " " + convert_to_string (chapter));

  // The chapter may have been added as it wasn't there before.
  // Check on that by checking whether the .svn directory is there.
  ustring chapterdirectory = project_data_directory_chapter (project, book, chapter);
  ustring dotsvndirectory = gw_build_filename (chapterdirectory, ".svn");
  bool chapterexisted = g_file_test (dotsvndirectory.c_str(), G_FILE_TEST_IS_DIR);
  if (!chapterexisted) {
    // Not there, add it. First add book, then add chapter.
    {
      ustring workingdirectory = project_data_directory_project (project);
      GwSpawn spawn ("svn");
      spawn.workingdirectory (workingdirectory);
      spawn.arg ("add");
      spawn.arg (books_id_to_english (book));
      spawn.devnull ();
      spawn.run ();
    }
    {
      ustring workingdirectory = project_data_directory_book (project, book);
      GwSpawn spawn ("svn");
      spawn.workingdirectory (workingdirectory);
      spawn.arg ("add");
      spawn.arg (convert_to_string (chapter));
      spawn.run ();
    }
  }

  // If the file existed, do an update, merging possible changes from others 
  // into the data. And then call other tasks as a result of this update.
  if (chapterexisted) {
    bool dummy;
    SubversionUpdateType subversionupdatetype = subversion_update_chapter_remote (project, book, chapter, dummy);
    switch (subversionupdatetype) {
      case sutOk:
      case sutError:
      case sutUpdated:
      case sutAdded:
      case sutReplaced:
      case sutMerged:
      case sutDeleted:
      {
        break;
      }
      case sutConflict:
      {
        subversion_resolve_conflict_chapter (project, book, chapter);
        break;
      }
    }
  }
  
  // Commit the changes.
  bool task_was_done;
  {
    GwSpawn spawn ("svn");
    spawn.workingdirectory (project_data_directory_project (project));
    spawn.arg ("commit");
    spawn.arg ("-m");
    spawn.arg ("Store chapter");
    spawn.arg ("--non-interactive");
    spawn.run ();
    task_was_done = (spawn.exitstatus == 0);
  }    
  
  // Return whether ok.
  return task_was_done;
}


Reference subversion_execute_retrieve_reference ()
// Retrieves the editor's reference from the database.
{
  Reference reference (0);
  extern InterprocessCommunication * ipc;
  vector <ustring> payload = ipc->get_payload (ipcctSubversionReference);
  if (payload.size() == 2) {
    reference.book = convert_to_int (payload[0]);
    reference.chapter = convert_to_int (payload[1]);
  }
  return reference;
}


bool subversion_execute_initialize_project_local (const ustring& project)
// Imports a project into the respository for the first time.
// Some important theory:
// The directory structure in the repository. Each chapter has its own
// subdirectory facilitating fast storage of a chapter.
{
  // Some important theory:
  // The directory structure in the repository. Each chapter has its own
  // subdirectory facilitating fast storage of a chapter.

  // Log
  gw_message ("Initialize project " + project);
  
  // Whether operation was ok.
  bool was_okay = true;
  
  // Directories.
  ustring workingdirectory = gw_build_filename (directories_get_projects(), project);

  // Import the project into the local repository.
  // (Remote repository is not initialized here, but in the Subversion setup dialog).
  if (was_okay) {
    GwSpawn spawn ("svn");
    spawn.workingdirectory (workingdirectory);
    spawn.arg ("import");
    spawn.arg ("data");
    spawn.arg (subversion_repository_local (project, true));
    spawn.arg ("-m");
    spawn.arg ("Initial import project");
    spawn.arg ("--non-interactive");
    spawn.run ();
    was_okay = (spawn.exitstatus == 0);
  }

  // Check the project out so our data changes into a working copy.
  // Save our data in case things go wrong.
  if (was_okay) {
    ustring datadirectory = project_data_directory_project (project);
    ustring checkoutdirectory = gw_build_filename (workingdirectory, gw_path_get_basename (subversion_repository_local (project, true)));

    GwSpawn spawn ("svn");
    spawn.workingdirectory (workingdirectory);
    spawn.arg ("checkout");
    spawn.arg (subversion_repository_local (project, true));
    spawn.arg ("--non-interactive");
    spawn.run ();
    was_okay = (spawn.exitstatus == 0);

    if (was_okay) {
      unix_rmdir (datadirectory);
      unix_mv (checkoutdirectory, datadirectory);
    } else {
      unix_rmdir (checkoutdirectory);
    }
  }
  
  // Return whether okay.
  return was_okay;
}


bool subversion_execute_commit_project (const ustring& project)
// Commits a project.
{
  ustring directory = project_data_directory_project (project);
  GwSpawn spawn ("svn");
  spawn.workingdirectory (directory);
  spawn.arg ("commit");
  spawn.arg ("-m");
  spawn.arg ("Commit project");
  spawn.arg ("--non-interactive");
  spawn.run ();
  return (spawn.exitstatus == 0);
}


void subversion_store_re_open_project (const ustring& project)
// Sends the project to be opened again.
{
  // Send the project.
  extern InterprocessCommunication * ipc;
  vector <ustring> payload;
  payload.push_back (project);
  ipc->send (ipcstBibleditBin, ipcctSubversionReopenProject, payload);
}


ustring subversion_retrieve_re_open_project ()
// Retrieves from the database whether the project should be re-opened.
{
  extern InterprocessCommunication * ipc;
  vector <ustring> payload = ipc->get_payload (ipcctSubversionReopenProject);
  ipc->erase_payload (ipcctSubversionReopenProject);
  ustring project;
  if (!payload.empty()) project = payload[0];
  return project;
}


void subversion_command_pause (bool pause)
// Sets a flag whether the subversion subsystem has to pause or resume.
{
  extern Settings * settings;
  settings->session.subversion_pause = pause;
}


ustring subversion_mine_conflict_marker ()
{
  ustring marker ("<<<<<<< .mine");
  return marker;
}


unsigned int subversion_tasks_count ()
{
  return subversiontasks.size();
}


vector <ustring> subversion_get_next_task ()
{
  vector <ustring> task;
  if (!subversiontasks.empty()) {
    task.push_back (convert_to_string (subversiontasks[0].task));
    task.push_back (                   subversiontasks[0].project);
    task.push_back (convert_to_string (subversiontasks[0].book));
    task.push_back (convert_to_string (subversiontasks[0].chapter));
  }
  return task;
}


void subversion_erase_task_done ()
{
  // Erase the task on top of the stack, but also all other equal tasks.
  if (!subversiontasks.empty()) {
    SubversionTask task = subversiontasks[0];
    vector <SubversionTask> newtasks;
    for (unsigned int i = 0; i < subversiontasks.size(); i++) {
      if (   task.task != subversiontasks[i].task
          || task.project != subversiontasks[i].project
          || task.book != subversiontasks[i].book
          || task.chapter != subversiontasks[i].chapter)
        newtasks.push_back (subversiontasks[i]);  
    }
    subversiontasks = newtasks;
  }
}


void subversion_fail_task_done ()
{
  if (!subversiontasks.empty()) {
    SubversionTask task = subversiontasks[0];
    subversion_erase_task_done ();
    task.failures++;
    subversiontasks.push_back (task);
  }
}


void subversion_erase_task (SubversionTaskType task, const ustring& project, unsigned int book, unsigned int chapter)
{
  vector <SubversionTask> newtasks;
  for (unsigned int i = 0; i < subversiontasks.size(); i++) {
    if (   task    != subversiontasks[i].task
        || project != subversiontasks[i].project
        || book    != subversiontasks[i].book
        || chapter != subversiontasks[i].chapter)
      newtasks.push_back (subversiontasks[i]);  
  }
  subversiontasks = newtasks;
}


void subversion_command_store_reference (const Reference& reference)
{
  extern InterprocessCommunication * ipc;
  vector <ustring> payload;
  payload.push_back (convert_to_string (reference.book));
  payload.push_back (convert_to_string (reference.chapter));
  ipc->send (ipcstBibleditSubversion, ipcctSubversionReference, payload);
}
