/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "table.h"
#include <libxml/xmlwriter.h>
#include "usfmtools.h"


TableCell::TableCell (const ustring& marker_in, const ustring& text_in)
{
  marker = marker_in;
  text = text_in;
}


TableRow::TableRow (int dummy)
{
}


XslFoTable::XslFoTable (Usfm * usfm, vector <ustring> * xslfolines)
{
  myusfm = usfm;
  myxslfolines = xslfolines;
}


XslFoTable::~XslFoTable ()
{
}


void XslFoTable::buffer (const ustring& marker, const ustring& text)
{
  buffered_markers.push_back (marker);
  buffered_texts.push_back (text);
}


bool XslFoTable::pending ()
{
  return !buffered_markers.empty();
}


void XslFoTable::process ()
{
  // Bail out if there is nothing to do.
  if (!pending ()) return;

  // Pass 1: Assemble the table rows and cells.
  
  // Raw table data storage.
  vector <TableRow> rows_raw;
  
  // Storage for one row.
  TableRow row (0);
  
  // Storage for retrieving the cell's content. We use a special algorithm,
  // so that stray data is picked up as well rather than discarded.
  ustring text_of_this_cell;
  
  // Go through the buffer.
  for (unsigned int i = 0; i < buffered_markers.size(); i++) {
    
    // Get all markers in this line, and keep the table-related ones.
    vector <ustring> markers = usfm_get_all_markers (buffered_texts[i]);
    markers.insert (markers.begin(), buffered_markers[i]);
    vector <ustring> relevant_markers;
    for (unsigned int i2 = 0; i2 < markers.size(); i2++) {
      if (myusfm->is_table_element (markers[i2])) {
        relevant_markers.push_back (markers[i2]);
      }
    }
    
    // Go through the relevant markers.
    for (unsigned int i2 = 0; i2 < relevant_markers.size(); i2++) {

      // Get the bit of text between this marker and the next one.
      ustring this_marker = usfm_get_full_opening_marker (relevant_markers[i2]);
      ustring next_marker = "\non_existing_marker";
      if (i2 + 1 < relevant_markers.size())
        next_marker = usfm_get_full_opening_marker (relevant_markers[i2 + 1]);
      size_t position = buffered_texts[i].find (this_marker);
      if (position != string::npos) {
        buffered_texts[i].erase (position, this_marker.length());
      }
      position = buffered_texts[i].find (next_marker);
      if (position != string::npos) {
        text_of_this_cell.append (buffered_texts[i].substr (0, position));
        buffered_texts[i].erase (0, position);
      } else {
        text_of_this_cell.append (buffered_texts[i]);
        buffered_texts[i].clear ();
      }

      // Handle the type of this table element.
      TableElementType element_type = myusfm->table_element_get_subtype (relevant_markers[i2]);
      switch (element_type) {
        case tetRow:
        {
          if (!row.cells.empty()) rows_raw.push_back (row);
          row.cells.clear ();
          break;
        }
        case tetHeading:
        case tetCell:
        {
          TableCell cell (relevant_markers[i2], text_of_this_cell);
          row.cells.push_back (cell);          
          text_of_this_cell.clear ();
          break;
        }
      }      
    }
  }

  // Save any data in the row.
  if (!row.cells.empty()) rows_raw.push_back (row);
    
  // Clear buffers.
  buffered_markers.clear ();
  buffered_texts.clear ();
  
  // Pass 2: Retrieve the highest column number in this table.
  unsigned int highest_column = 0;
  for (unsigned int r = 0; r < rows_raw.size(); r++) {
    for (unsigned int c = 0; c < rows_raw[r].cells.size(); c++) {
      myusfm->is_table_element (rows_raw[r].cells[c].marker);
      if (myusfm->userint1 () > (int) highest_column)
        highest_column = myusfm->userint1 ();
    }
  }

  // Pass 3: Produce a tidied table.
  // This table has the same amount of columns in each row.
  // It may have empty cells in cases there was no data for it.
  vector <TableRow> rows_tidy;
  for (unsigned int r = 0; r < rows_raw.size(); r++) {
    TableRow row (0);
    for (unsigned int c = 0; c < highest_column; c++) {
      TableCell cell ("", "");
      row.cells.push_back (cell);
    }
    for (unsigned int c = 0; c < rows_raw[r].cells.size(); c++) {
      myusfm->is_table_element (rows_raw[r].cells[c].marker);
      int column = myusfm->userint1 ();
      if (column < 1) column = 1;
      column--;
      row.cells[column].marker = rows_raw[r].cells[c].marker;
      // Append text to any already existing tetxt in the cell. Very unlikely to happen...
      ustring text = row.cells[column].text;
      if (!text.empty ()) text.append (" ");
      text.append (rows_raw[r].cells[c].text);
      row.cells[column].text = text;      
    }
    rows_tidy.push_back (row);
  }  

  // Pass 4: Produce the xslfo code for this table.
  
  // Start the xml bit for the table to be created.
  xmlBufferPtr buffer = xmlBufferCreate();
  xmlTextWriterPtr writer = xmlNewTextWriterMemory(buffer, 0);
  xmlTextWriterSetIndent(writer, 1);
  xmlTextWriterStartElement (writer, BAD_CAST "fo:table");

  // Go through each row.
  for (unsigned int r = 0; r < rows_tidy.size(); r++) {
    TableRow row = rows_tidy[r];
    if (row.cells.empty ()) continue;
      
    // If the first row is a header, handle it, else it is the table body.
    ustring header_or_body = "fo:table-";
    bool heading = false;
    if (r == 0) {
      TableElementType element_type = myusfm->table_element_get_subtype (row.cells[0].marker);
      if (element_type == tetHeading) {
        heading = true;
      }
    }
    if (heading) header_or_body.append ("header");
    else header_or_body.append ("body");
    xmlTextWriterStartElement (writer, BAD_CAST header_or_body.c_str());
    xmlTextWriterStartElement (writer, BAD_CAST "fo:table-row");

    // Handle each cell.
    for (unsigned int c = 0; c < row.cells.size(); c++) {
      xmlTextWriterStartElement (writer, BAD_CAST "fo:table-cell");
      /*
      xmlTextWriterStartElement (writer, BAD_CAST "fo:block");
      xmlTextWriterWriteFormatString (writer, "%s", row.cells[c].text.c_str());
      xmlTextWriterEndElement (writer);
      */
      vector <ustring> lines;
      {
        ustring marker (row.cells[c].marker);
        if (marker.empty ()) marker = "tc1";
        myusfm->is_table_element (marker);
        XmlFoBlock block (&lines, 
                          myusfm->fontsize(), 100, myusfm->italic(), myusfm->bold(), myusfm->underline(), 
                          myusfm->smallcaps(), myusfm->justification(),
                          myusfm->spacebefore(), myusfm->spaceafter(), myusfm->leftmargin(),
                          myusfm->rightmargin(), myusfm->firstlineindent(), false,
                          false);
        lines.push_back (row.cells[c].text);
      }
      for (unsigned int i = 0; i < lines.size(); i++) {
        xmlTextWriterWriteRaw (writer, BAD_CAST (lines[i].c_str()));
      }
      /*
      */
      xmlTextWriterEndElement (writer);      
    }
      
    // Close row.  
    xmlTextWriterEndElement (writer);      
    xmlTextWriterEndElement (writer);      
  }

  // Close document and write it to the xslfo lines.
  xmlTextWriterEndDocument (writer);
  xmlTextWriterFlush (writer);
  myxslfolines->push_back ((const gchar *) buffer->content);

  // Free memory.
  if (writer) xmlFreeTextWriter (writer);
  if (buffer) xmlBufferFree (buffer);

}
