/*
** Copyright (C) 2003-2007 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 3 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "usfm2abiword.h"
#include "utilities.h"
#include "constants.h"
#include "abiwordutils.h"
#include "color.h"
#include "usfmtools.h"
#include "xmlutils.h"
#include "gwrappers.h"
#include "progresswindow.h"


Usfm2Abiword::Usfm2Abiword (const vector<ustring>& lines, vector<Style> * styles)
{
  // Save or initialize variables.
  mystyles = styles;
  font = "Bitstream Charter";
  footnoteid = 0;
  endnoteid = 0;
  success = true;
  xid = 0;
  
  // Create a new XML buffer, to which the document will be written
  buffer = xmlBufferCreate();
  if (!buffer) {
    gw_critical ("Error creating the xml buffer");
    success = false;
    return;
  }
  
  // Create a new XmlWriter for memory, with no compression.
  writer = xmlNewTextWriterMemory(buffer, 0);
  if (!writer) {
    gw_critical ("Error creating the xml writer");
    success = false;
    return;
  }

  // Indent the output.
  xmlTextWriterSetIndent(writer, 1);
  
  // Create first bit of the .abw document.
  int rc;
  rc = xmlTextWriterStartDocument(writer, NULL, "UTF-8", NULL);
  if (rc < 0) {
    gw_critical ("Error starting the xml document");
  }
  
  // Write DTD.
  xmlTextWriterWriteDTD (writer, BAD_CAST "abiword", BAD_CAST "-//ABISOURCE//DTD AWML 1.0 Strict//EN", BAD_CAST "http://www.abisource.com/awml.dtd", NULL);

  // Abiword file. Leave abiword element open.
  xmlTextWriterStartElement (writer, BAD_CAST "abiword");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "template", BAD_CAST "false");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "styles", BAD_CAST "unlocked");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns:fo", BAD_CAST "http://www.w3.org/1999/XSL/Format");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns:math", BAD_CAST "http://www.w3.org/1998/Math/MathML");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xid-max", BAD_CAST "999999");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns:dc", BAD_CAST "http://purl.org/dc/elements/1.1/");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "fileformat", BAD_CAST "1.0");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns:svg", BAD_CAST "http://www.w3.org/2000/svg");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns:awml", BAD_CAST "http://www.abisource.com/awml.dtd");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns", BAD_CAST "http://www.abisource.com/awml.dtd");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xmlns:xlink", BAD_CAST "http://www.w3.org/1999/xlink");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "version", BAD_CAST "0.99.2");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "xml:space", BAD_CAST "preserve");
  xmlTextWriterWriteAttribute(writer, BAD_CAST "props", BAD_CAST "dom-dir:ltr; document-footnote-restart-section:0; document-endnote-type:numeric; document-endnote-place-enddoc:1; document-endnote-initial:1; lang:en-US; document-endnote-restart-section:0; document-footnote-restart-page:0; document-footnote-type:numeric; document-footnote-initial:1; document-endnote-place-endsection:0");
  
  // Meta data.
  xmlTextWriterStartElement (writer, BAD_CAST "metadata");
  xmlTextWriterStartElement (writer, BAD_CAST "m");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "key", BAD_CAST "dc.format");
  xmlTextWriterWriteFormatString (writer, "application/x-abiword");
  xmlTextWriterEndElement (writer);
  xmlTextWriterStartElement (writer, BAD_CAST "m");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "key", BAD_CAST "abiword.generator");
  xmlTextWriterWriteFormatString (writer, "AbiWord");
  xmlTextWriterEndElement (writer);
  xmlTextWriterEndElement (writer);
  
  // Styles.
  assemble_styles ();

  // Page.  
  xmlTextWriterStartElement (writer, BAD_CAST "pagesize");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "pagetype", BAD_CAST "A4");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "orientation", BAD_CAST "portrait");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "width", BAD_CAST "210.000000");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "height", BAD_CAST "297.000000");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "units", BAD_CAST "mm");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "page-scale", BAD_CAST "1.000000");
  xmlTextWriterEndElement (writer);
  
  // Open the section section.
  xmlTextWriterStartElement (writer, BAD_CAST "section");
  write_xid_property ();
  xmlTextWriterWriteAttribute (writer, BAD_CAST "props", BAD_CAST "page-margin-footer:0.5in; page-margin-header:0.5in");
  
  // Go through all the lines and make full paragraphs out of them.
  vector <ustring> paragraphs;
  collect_paragraphs (lines, paragraphs);  
  
  // Process all paragraphs.
  for (unsigned int i = 0; i < paragraphs.size(); i++) {
    write_paragraph (paragraphs[i]);
  }
  
  // We could close any open elements by hand, but closing the document does it all.
  xmlTextWriterEndDocument (writer);
  
  // Flush the writer, else nothing gets transferred to the buffer.
  xmlTextWriterFlush (writer);

  // Make lines of text out of the buffer.
  ParseLine parseline ((const char *) buffer->content);
  output = parseline.lines;  
}


Usfm2Abiword::~Usfm2Abiword ()
{
  // Free some memory.
  if (writer) xmlFreeTextWriter (writer);
  if (buffer) xmlBufferFree (buffer);
}


void Usfm2Abiword::write_paragraph (ustring paragraph)
{
  // Extract the marker of the paragraph.
  ustring line_without_first_marker (paragraph);
  ustring marker = usfm_extract_marker (line_without_first_marker);
  
  // Find a Style for that marker, or NULL if it isn't there.
  Style * style = NULL;
  for (unsigned int i = 0; i < mystyles->size(); i++) {
    if (mystyles->at(i).marker == marker) {
      style = &mystyles->at(i);
      break;
    }
  }

  // Based on the above information, produce a style marker.
  ustring style_marker;
  if (style) style_marker = style->marker;

  // Obtain the type and the subtype.
  StyleType style_type = (StyleType) G_MAXINT;
  int style_subtype = -1;
  if (style) {
    style_type = style->type;
    style_subtype = style->subtype;
  }
  
  // Produce flags whether it is a paragraph, and so forth.
  AbiwordStyleType abiwordstyletype;
  bool plaintext;
  bool showmarker;
  abiword_style_get_properties (style_type, style_subtype, abiwordstyletype, plaintext, showmarker);

  // Open paragraph.
  xmlTextWriterStartElement (writer, BAD_CAST "p");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "style", BAD_CAST abiword_style_produce_name (style_marker).c_str());
  write_xid_property ();
  
  // Produce the line of text.
  ustring textline;
  if (showmarker) textline = paragraph;
  else textline = line_without_first_marker;
    
  // Change some characters to xml entities.
  xml_handle_entities (textline, NULL);

  /*
  The processing of inline text and of notes needs to be improved a bit.
  Right now we are writing raw xml code.
  But we need to use libxml2's function for writing.
  This can be done if it is done bit by bit.
  E.g. the first inline marker or footnote marker is seen, then the next
  one is looked for, and that bit of text snooped off the line.
  Then xml writing functions are called, and that line is written.
  Then we continue with the next inline or footnote marker, and so on.
  This also has the advantage that it is much safer, as we won't have
  overlapping styles.
  This implies that the inline and footnote processing should be unified in 
  one function.
  We then will have a streaming operation on all markers we know of.
  */
  
  // Inline text processing.
  process_inline_text (textline); 

  // Footnote and endnote processing.
  process_notes (textline);  
  
  // Add this processed line to the output.
  if (!textline.empty()) {
    xmlTextWriterWriteRaw (writer, BAD_CAST textline.c_str());
  }
  
  // Close paragraph.
  xmlTextWriterEndElement (writer);
}


void Usfm2Abiword::add_xid (ustring& line)
{
  xid++;
  line.append (" xid=\"" + convert_to_string (xid) + "\"");
}


void Usfm2Abiword::write_xid_property ()
{
  xid++;
  xmlTextWriterWriteAttribute (writer, BAD_CAST "xid", BAD_CAST convert_to_string (xid).c_str());
}


void Usfm2Abiword::assemble_styles ()
/*
Based on the available styles, it produces the lines that go into the .abw document.
*/
{
  // Open the styles element.
  xmlTextWriterStartElement (writer, BAD_CAST "styles");

  // Write the style for a marker that is not in the stylesheet.
  Style style ("", "", false);
  style.name = abiword_unknown_style ();
  assemble_style (&style, true, true, "");
  
  // Assemble the style(s) for the markers that are in the stylesheet.
  for (unsigned int i = 0; i < mystyles->size(); i++) {
    AbiwordStyleType abiwordstyletype;
    bool plaintext, showmarker;
    abiword_style_get_properties (mystyles->at(i).type, mystyles->at(i).subtype, abiwordstyletype, plaintext, showmarker);
    assemble_style (&mystyles->at(i), abiwordstyletype == abwstParagraph, plaintext, font);
  }

  // Assemble styles for foot- and endnote.
  assemble_notes_styles ();  
  
  // End the element.
  xmlTextWriterEndElement (writer);
}


void Usfm2Abiword::assemble_style (Style * style, bool paragraph, bool plaintext, const ustring& font)
/*
This assembles a style for an .abw document.

Input:
- style: the style to convert.
- paragraph: True: This is a paragraph style. False: This is a character style
- plaintext: If this is true, then no style information is used, but it is shown as plain text.
- font: Only used for paragraph styles. It is the font to use.
  
Example paragraph styles:
"<s type=\"P\" name=\"Normal\" followedby=\"Current Settings\" props=\"font-family:Times New Roman; margin-top:0pt; font-variant:normal; margin-left:0pt; widows:2; font-style:normal; text-indent:0in; text-position:normal; line-height:1.0; margin-bottom:0pt; bgcolor:transparent; margin-right:0pt; font-size:12pt; color:000000; text-decoration:none; font-weight:normal; text-align:left; font-stretch:normal\"/>"
<s type="P" name="Mystyle (xxx)" basedon="None" followedby="Current Settings" props="text-folded-id:0; lang:en-US; font-variant:normal; margin-bottom:20pt; color:000000; font-size:16pt; margin-right:6.0mm; text-position:normal; bgcolor:transparent; height:0in; format:%*%d.; font-style:italic; list-delim:%L; text-decoration:none; keep-together:no; text-align:left; widows:0; list-tag:0; keep-with-next:no; text-indent:0in; width:0in; line-height:1.0; start-value:1; list-decimal:.; orphans:0; font-weight:normal; margin-left:5.0mm; text-folded:0; font-stretch:normal; font-family:DejaVu Sans Light; default-tab-interval:0.5in; list-style:None; homogeneous:1; display:inline; dom-dir:ltr; margin-top:10pt"/>

Example character style:
<s type="C" name="test" basedon="None" followedby="Current Settings" props="text-indent:0in; orphans:2; margin-top:0pt; margin-left:0pt; list-decimal:.; line-height:1.0; keep-with-next:no; lang:en-US; text-align:left; list-style:None; keep-together:no; color:000000; dom-dir:ltr; margin-bottom:0pt; font-weight:bold; text-decoration:none; bgcolor:transparent; start-value:1; font-stretch:normal; font-variant:normal; font-size:12pt; margin-right:0pt; font-style:italic; widows:2; list-delim:%L; font-family:DejaVu Sans"/>
*/
{
  xmlTextWriterStartElement (writer, BAD_CAST "s");
  xmlChar * xmlchar;
  if (paragraph) 
    xmlchar = xmlStrdup (BAD_CAST "P");
  else 
    xmlchar = xmlStrdup (BAD_CAST "C");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "type", xmlchar);
  xmlFree (xmlchar);

  xmlTextWriterWriteAttribute (writer, BAD_CAST "name", BAD_CAST abiword_style_produce_name (style->marker).c_str());

  xmlTextWriterWriteAttribute (writer, BAD_CAST "followedby", BAD_CAST "Current Settings");

  // Produce the properties.
  ustring properties;
  bool first_property_added = false;
 
  ustring myfont (font);
  if (plaintext) myfont = "Courier";
  if (paragraph || plaintext)
    abiword_style_add_property (first_property_added, properties, "font-family", myfont, "");
  if (!plaintext) {

    double fontsize = style->fontsize;
    if (!paragraph) {
      if (style->fontpercentage != 100)
        fontsize = 12 * style->fontpercentage / 100;
    } 
    if (paragraph || (style->fontpercentage != 100)) {
      if (!style->superscript) {
        abiword_style_add_property (first_property_added, properties, "font-size", convert_to_string (fontsize), "pt");
      }
    }
    
    // Italic, bold, underline, smallcaps can be ON, OFF, INHERIT, or TOGGLE.
    // Not all properties work in abiword yet.
    if ((style->italic == ON) || (style->italic == TOGGLE)) 
      abiword_style_add_property (first_property_added, properties, "font-style", "italic", "");
    if ((style->bold == ON) || (style->bold == TOGGLE)) 
      abiword_style_add_property (first_property_added, properties, "font-weight", "bold", "");
    if ((style->underline == ON) || (style->underline == TOGGLE)) 
      abiword_style_add_property (first_property_added, properties, "text-decoration", "underline", "");
  
    // As of April 2007, Small Caps has not yet been implemented in Abiword.
    // See http://bugzilla.abisource.com/show_bug.cgi?id=1851
    // Not sure which of the following two is going to work, so implementing both of them...
    if ((style->smallcaps == ON) || (style->smallcaps == TOGGLE)) {
      abiword_style_add_property (first_property_added, properties, "text-transform", "small-caps", "");
      abiword_style_add_property (first_property_added, properties, "font-variant", "small-caps", "");
    }
  
    if (style->superscript) 
      abiword_style_add_property (first_property_added, properties, "text-position", "superscript", "");

    // Styles that only occur in paragraphs, not in character styles.  
    if (paragraph) {
      
      ustring text_align;  
      if (style->justification == CENTER) {
        text_align.append ("center");
      } else if (style->justification == RIGHT) {
        text_align.append ("right");
      } else if (style->justification == JUSTIFIED) {
        text_align.append ("justify");
      } else { // Default is LEFT.
        text_align.append ("left");
      }
      abiword_style_add_property (first_property_added, properties, "text-align", text_align, "");

      if (style->spacebefore != 0)
        abiword_style_add_property (first_property_added, properties, "margin-top", convert_to_string (style->spacebefore), "mm");
      if (style->spaceafter != 0)
        abiword_style_add_property (first_property_added, properties, "margin-bottom", convert_to_string (style->spaceafter), "mm");

      if (style->leftmargin != 0)
        abiword_style_add_property (first_property_added, properties, "margin-left", convert_to_string (style->leftmargin), "mm");
      if (style->rightmargin != 0)
        abiword_style_add_property (first_property_added, properties, "margin-right", convert_to_string (style->rightmargin), "mm");
      if (style->firstlineindent != 0)
        abiword_style_add_property (first_property_added, properties, "text-indent", convert_to_string (style->firstlineindent), "mm");
   
    }

    if (style->color)
      abiword_style_add_property (first_property_added, properties, "color", color_decimal_to_hex (style->color), "");

  }

  // Add any properties that are there. 
  if (!properties.empty()) {
    xmlTextWriterWriteAttribute (writer, BAD_CAST "props", BAD_CAST properties.c_str());
  }
  
  // Close style.
  xmlTextWriterEndElement (writer);
}


void Usfm2Abiword::assemble_notes_styles ()
/*
This function assenbles the styles for the footnote and the endnote.

Here's how Abiword writes the styles:

<s type="C" name="Footnote Reference" basedon="None" followedby="Current Settings" props="text-position:superscript; font-size:10pt"/>
<s type="C" name="Footnote Text" basedon="Normal" followedby="Current Settings" props="text-position:normal; font-size:10pt"/>
<s type="C" name="Endnote Reference" basedon="None" followedby="Current Settings" props="text-position:superscript; font-size:10pt"/>
<s type="C" name="Endnote Text" basedon="Normal" followedby="Current Settings" props="text-position:normal"/>

*/
{
  xmlTextWriterStartElement (writer, BAD_CAST "s");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "type", BAD_CAST "C");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "name", BAD_CAST "Footnote Reference");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "basedon", BAD_CAST "None");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "followedby", BAD_CAST "Current Settings");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "props", BAD_CAST "text-position:superscript; font-size:10pt");
  xmlTextWriterEndElement (writer);
  
  xmlTextWriterStartElement (writer, BAD_CAST "s");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "type", BAD_CAST "C");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "name", BAD_CAST "Footnote Text");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "basedon", BAD_CAST "Normal");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "followedby", BAD_CAST "Current Settings");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "props", BAD_CAST "text-position:normal");
  xmlTextWriterEndElement (writer);
  
  xmlTextWriterStartElement (writer, BAD_CAST "s");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "type", BAD_CAST "C");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "name", BAD_CAST "Endnote Reference");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "basedon", BAD_CAST "None");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "followedby", BAD_CAST "Current Settings");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "props", BAD_CAST "text-position:superscript; font-size:10pt");
  xmlTextWriterEndElement (writer);
  
  xmlTextWriterStartElement (writer, BAD_CAST "s");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "type", BAD_CAST "C");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "name", BAD_CAST "Endnote Text");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "basedon", BAD_CAST "None");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "followedby", BAD_CAST "Current Settings");
  xmlTextWriterWriteAttribute (writer, BAD_CAST "props", BAD_CAST "text-position:normal");
  xmlTextWriterEndElement (writer);
}


void Usfm2Abiword::collect_paragraphs (const vector<ustring>& lines, vector<ustring>& paragraphs)
/*
This function goes through all the "lines", and assembles paragraphs
out of them, and stores that whole paragraph as one line in "paragraphs",
and then goes on with the next, and so on, till all paragraphs have been done.
This function presumes that the "lines" it gets have been cleaned already,
in the sense that each line either starts a new paragraph or a new verse.
*/
{
  // Storage for the paragraph.
  ustring paragraph;
  
  // Go through all the lines.
  for (unsigned int i = 0; i < lines.size(); i++) {
    
    // Get the marker of the line.
    ustring s (lines[i]);
    ustring marker (usfm_extract_marker (s));
    
    // Find the type and subtype of the marker.
    StyleType style_type = (StyleType) G_MAXINT;
    int style_subtype = G_MAXINT;
    for (unsigned int i2 = 0; i2 < mystyles->size(); i2++) {
      if (mystyles->at(i2).marker == marker) {
        style_type = mystyles->at(i2).type;
        style_subtype = mystyles->at(i2).subtype;
        break;
      }
    }
    
    // Produce flags whether it is a paragraph, or something else.
    AbiwordStyleType abiwordstyletype;
    bool b;
    abiword_style_get_properties (style_type, style_subtype, abiwordstyletype, b, b);

    // If this line starts a new paragraph, process the old paragraph first.
    if ((abiwordstyletype == abwstParagraph) && ((!paragraph.empty()))) {
      paragraphs.push_back (paragraph);
      paragraph.clear ();
    }

    // Add this data to the paragraph.
    if (!paragraph.empty()) paragraph.append (" ");
    paragraph.append (lines[i]);    
  }
  
  // If there is something in the last paragraph, process that also.
  if (!paragraph.empty())
    paragraphs.push_back (paragraph);
}


void Usfm2Abiword::process_inline_text (ustring& paragraph)
/*
This function takes a paragraph, and looks for all the inline styles that it
has. It then generates all the full opening and full closing markesr that
belong to this style. Next it generates the Abiword code that would replace
those markers. Example of such code:
<p style="Normal, First Line Indent p" xid="14">\v 1 The <c style="* Name of Deity nd">Lord</c> told Moses.</p>
*/ 
{
  // Get a container with all the unique markers in the line.
  vector<ustring> markers;
  {
    markers = usfm_get_all_markers (paragraph);
    set<ustring> markers_set (markers.begin(), markers.end());
    markers.clear ();
    markers.assign (markers_set.begin(), markers_set.end());
  }
  
  // Go through all markers found.
  for (unsigned int i = 0; i < markers.size(); i++) {

    // Find the type and subtype of the marker, and the name.
    StyleType style_type = (StyleType) G_MAXINT;
    int style_subtype = G_MAXINT;
    for (unsigned int i2 = 0; i2 < mystyles->size(); i2++) {
      if (markers[i] == mystyles->at(i2).marker) {
        style_type = mystyles->at(i2).type;
        style_subtype = mystyles->at(i2).subtype;
        break;
      }
    }
    
    // Produce flags whether it is inline text, or a verse.
    AbiwordStyleType abiwordstyletype;
    bool b;
    abiword_style_get_properties (style_type, style_subtype, abiwordstyletype, b, b);
    
    // Store data for a character style, that is, inline text, and for a verse style
    if ((abiwordstyletype == abwstCharacter) || (abiwordstyletype == abwstVerse)) {
      ustring opener = usfm_get_full_opening_marker (markers[i]);
      ustring closer = usfm_get_full_closing_marker (markers[i]);
      ustring opening_code = "<c style=\"" + abiword_style_produce_name (markers[i]) + "\">";
      ustring closing_code = "</c>";
      if (abiwordstyletype == abwstCharacter) {
        place_styles_for_inline_text (paragraph, opener, closer, opening_code, closing_code);
      }
      if (abiwordstyletype == abwstVerse) {
        place_styles_for_verse (paragraph, opener, opening_code, closing_code);
      }
    }
  }
}


void Usfm2Abiword::place_styles_for_verse (ustring& paragraph, const ustring& verse_marker, const ustring& style_opener, const ustring& style_closer)
/*
This replaces amy "verse marker", with the appropriate abiword style code.
*/
{
  // Replace all verse markers with the opening code.
  replace_text (paragraph, verse_marker, style_opener);
  
  // Go through the paragraph looking for the verse style openers. 
  // There could be several of them in a paragraph.
  size_t position = paragraph.find (style_opener);
  while (position != string::npos) {
    // Find the position after the verse number, and put the closing code there.
    size_t end_of_verse_pos = paragraph.find (" ", position + style_opener.length());
    if (end_of_verse_pos != string::npos) {
      paragraph.insert (end_of_verse_pos, style_closer);
    } else {
      paragraph.append (style_closer);
    }
    // Next verse marker.
    position = paragraph.find (style_opener, ++position);    
  }
}


void Usfm2Abiword::place_styles_for_inline_text (ustring& paragraph, const ustring& opener, const ustring& closer, const ustring& style_opener, const ustring& style_closer)
/*
1. Usfm text that has the opening marker, but not the closing marker, may result
   in malformed output. To ensure this does not occur, the opening tag
   is only placed in the output when the closing usfm is found too.
2. For inline text markers, the space following the opening marker is
   part of the marker, and not part of the word.
3. In rare cases, while comparing, it occurred that the closing marker comes
   before the opening marker. Though both markers are in, they are not in the 
   right order, and so should not be transformed else an error occurs in 
   Abiword.

Variables:
paragraph    : the paragraph that contains the usfm text to be modified.
opener       : the opening marker.
closer       : the closing marker.
style_opener : the opening code.
style_closer : the closing code.
*/
{
  // Allow for a marker to occur more than once in the paragraph.
  bool still_working = true;
  size_t begin_position = 0;
  size_t end_position = 0;
  while (still_working) {
    still_working = false;
    // Search for a opener after the position of the previous closer.
    begin_position = paragraph.find(opener, end_position);
    if (begin_position != string::npos) {
      // Search for a closer after the position of the previous opener.
      end_position = paragraph.find(closer, begin_position);
      if (end_position != string::npos) {
        // Check that the closer comes after the opener.
        if (end_position > begin_position) {
          // At this stage we have both the opening and closing marker, 
          // and the closer comes after the opener, 
          // so we're safe to replace the markers with the abiword style tags.
          size_t position;
          size_t length;
          // Do higher position first, so changes made won't affect the lower position.
          position = end_position;
          length = closer.length();
          paragraph.replace (position, length, style_closer);
          // Do lower position next.          
          position = begin_position;
          length = opener.length();
          paragraph.replace (position, length, style_opener);
          // Next iteration.
          still_working = true;
        }
      }
    }
  }
}


void Usfm2Abiword::process_notes (ustring& paragraph)
// Find the usfm data to extract the notes, and call the transformation routines.
{
  // Go through all styles to look for the right markers that opens
  // a footnote, a crossreference, and an endnote.
  for (unsigned int i = 0; i < mystyles->size(); i++) {
    ustring marker;          // The marker
    bool abifootnote = true; // Whether footnote or endnote in abiword.
    if (mystyles->at(i).type == stFootEndNote) {
      if (mystyles->at(i).subtype == fentFootnote) {
        marker = mystyles->at(i).marker;
        abifootnote = true;
      }
      if (mystyles->at(i).subtype == fentEndnote) {
        marker = mystyles->at(i).marker;
        abifootnote = false;
      }
    }
    if (mystyles->at(i).type == stCrossreference) {
      if (mystyles->at(i).subtype == ctCrossreference) {
        marker = mystyles->at(i).marker;
        abifootnote = true;
      }
    }
    
    // If a marker has been found, process it.
    if (!marker.empty()) {
      // Get full openers and closers.
      ustring full_opener = usfm_get_full_opening_marker (marker);
      ustring full_closer = usfm_get_full_closing_marker (marker);
      // Process abiword footnote.
      if (abifootnote) {
        process_footnote (paragraph, full_opener, full_closer);
      } else {
        process_footnote (paragraph, full_opener, full_closer);
        // process_endnote (paragraph, full_opener, full_closer);
      }
    }
    
  }
}


void Usfm2Abiword::process_footnote (ustring& paragraph, const ustring& opener, const ustring& closer)
/*
Extract the footnotes and crossreferences. Format them for abiword.

Example of a footnote in Abiword:

<field footnote-id="0" type="footnote_ref" style="Footnote Reference" xid="4"></field><foot footnote-id="0" xid="1">
<p footnote-id="0" style="Footnote Text" xid="2"><field footnote-id="0" type="footnote_anchor" style="Footnote Text" xid="5" props="text-position:superscript"></field><c footnote-id="0"> footnote.</c></p>
</foot>

If there is more than one footnote, the footnote-id increases by one.

*/
{
  // Allow for a marker to occur more than once in the paragraph.
  bool still_working = true;
  size_t begin_position = 0;
  size_t end_position = 0;
  while (still_working) {
    still_working = false;
    // Search for a opener after the position of the previous closer.
    begin_position = paragraph.find(opener, end_position);
    if (begin_position != string::npos) {
      // Search for a closer after the position of the previous opener.
      end_position = paragraph.find(closer, begin_position);
      if (end_position != string::npos) {
        // Check that the closer comes after the opener.
        if (end_position > begin_position) {
          // At this stage we have both the opening and closing marker, 
          // and the closer comes after the opener, 
          // so we're safe to process this footnote.
          
          // Assemble the opening abiword code.
          ustring opening_code = "<field footnote-id=\"";
          opening_code.append (convert_to_string (footnoteid));
          opening_code.append ("\" type=\"footnote_ref\" style=\"Footnote Reference\"");
          add_xid (opening_code);
          opening_code.append ("></field><foot footnote-id=\"");
          opening_code.append (convert_to_string (footnoteid));
          opening_code.append ("\"");
          add_xid (opening_code);
          opening_code.append (">\n");
          
          opening_code.append ("<p footnote-id=\"");
          opening_code.append (convert_to_string (footnoteid));
          opening_code.append ("\" style=\"Footnote Text\"");
          add_xid (opening_code);
          opening_code.append ("><field footnote-id=\"");
          opening_code.append (convert_to_string (footnoteid));
          opening_code.append ("\" type=\"footnote_anchor\" style=\"Footnote Text\"");
          add_xid (opening_code);
          opening_code.append (" props=\"text-position:superscript\"></field><c footnote-id=\"");
          opening_code.append (convert_to_string (footnoteid));
          opening_code.append ("\"> ");
          
          // Assemble the closing abiword code.
          ustring closing_code = "</c></p>\n";
          closing_code.append ("</foot>");
          
          // Notice the change in length, for later usage.
          size_t length_change = paragraph.length ();
          
          // Insert the closing code first, so that the change of the length
          // of the paragraph does not affect the opening position.
          size_t position = end_position + closer.length();
          paragraph.insert (position, closing_code);
          // Insert opening code.
          position = begin_position;
          paragraph.insert (position, opening_code);
          
          // Calculate the change in length. Adapt the position to search from
          // for the next possible footnote.
          length_change = paragraph.length() - length_change;
          end_position += length_change;
          
          // Next footnote id.
          footnoteid++;
          
          // Next iteration.
          still_working = true;
        }
      }
    }
  }
}


void Usfm2Abiword::process_endnote (ustring& paragraph, const ustring& opener, const ustring& closer)
/*
Extract the endnotes, and format them for abiword.

Example of an endnote:

<field endnote-id="0" style="Endnote Reference" type="endnote_ref" xid="12"></field><endnote endnote-id="0" xid="9">
<p endnote-id="0" style="Endnote Text" xid="10"><field endnote-id="0" style="Endnote Text" type="endnote_anchor" xid="13" props="text-position:superscript"></field> Endnote</p>
</endnote>

If there is more than one endnote, the endnote-id increases by one.

*/
{
  /*
  
This code is not now used, but left just in case we wish to use it later on.
  
  // Allow for an endnote to occur more than once in the paragraph.
  bool still_working = true;
  size_t begin_position = 0;
  size_t end_position = 0;
  while (still_working) {
    still_working = false;
    // Search for a opener after the position of the previous closer.
    begin_position = paragraph.find(opener, end_position);
    if (begin_position != string::npos) {
      // Search for a closer after the position of the previous opener.
      end_position = paragraph.find(closer, begin_position);
      if (end_position != string::npos) {
        // Check that the closer comes after the opener.
        if (end_position > begin_position) {
          // At this stage we have both the opening and closing marker, 
          // and the closer comes after the opener, 
          // so we're safe to process this endnote.
          
          // Assemble the opening abiword code.
          ustring opening_code = "<field endnote-id=\"";
          opening_code.append (convert_to_string (endnoteid));
          opening_code.append ("\" type=\"endnote_ref\" style=\"Endnote Reference\"");
          add_xid (opening_code);
          opening_code.append ("></field><endnote endnote-id=\"");
          opening_code.append (convert_to_string (endnoteid));
          opening_code.append ("\"");
          add_xid (opening_code);
          opening_code.append (">\n");

          opening_code.append ("<p endnote-id=\"");
          opening_code.append (convert_to_string (endnoteid));
          opening_code.append ("\" style=\"Endnote Text\"");
          add_xid (opening_code);
          opening_code.append ("><field endnote-id=\"");
          opening_code.append (convert_to_string (endnoteid));
          opening_code.append ("\" type=\"endnote_anchor\" style=\"Endnote Text\"");
          add_xid (opening_code);
          opening_code.append (" props=\"text-position:superscript\"></field> ");

          // Assemble the closing abiword code.
          ustring closing_code = "</p>\n";
          closing_code.append ("</endnote>");
          
          // Notice the change in length, for later usage.
          size_t length_change = paragraph.length ();
          
          // Insert the closing code first, so that the change of the length
          // of the paragraph does not affect the opening position.
          size_t position = end_position + closer.length();
          paragraph.insert (position, closing_code);
          // Insert opening code.
          position = begin_position;
          paragraph.insert (position, opening_code);
          
          // Calculate the change in length. Adapt the position to search from
          // for the next possible endnote.
          length_change = paragraph.length() - length_change;
          end_position += length_change;
          
          // Next endnote id.
          endnoteid++;
          
          // Next iteration.
          still_working = true;
        }
      }
    }
  }
  */
}

// todo Implement the tables export to AbiWord.
